import aerosandbox.numpy as np
from aerosandbox.geometry import Airfoil
from aerosandbox.geometry.airfoil.airfoil_families import get_kulfan_coordinates
from scipy import optimize
import matplotlib.pyplot as plt
import aerosandbox.tools.pretty_plots as p

if __name__ == '__main__':

    ### Design Conditions
    Re_des = 3e5  # Re to design to
    Cl_start = 1.0  # Lower bound of CLs that you care about
    Cl_end = 1.5  # Upper bound of CLs that you care about (Effectively, CL_max)
    Cm_min = -0.08  # Worst-allowable pitching moment that you'll allow
    TE_thickness = 0.0015  # Sets trailing edge thickness
    enforce_continuous_LE_radius = True  # Should we force the leading edge to have continous curvature?

    ### Guesses for airfoil CST parameters; you usually don't need to change these
    lower_guess = -0.05 * np.ones(30)
    upper_guess = 0.25 * np.ones(30)
    upper_guess[0] = 0.15
    upper_guess[1] = 0.20

    # lower_guess = [-0.21178419, -0.05500152, -0.04540216, -0.03436429, -0.03305599,
    #                -0.03121454, -0.04513736, -0.05491045, -0.02861083, -0.05673649,
    #                -0.06402239, -0.05963394, -0.0417384, -0.0310728, -0.04983729,
    #                -0.04211283, -0.04999657, -0.0632682, -0.07226548, -0.03604782,
    #                -0.06151112, -0.04030985, -0.02748867, -0.02705322, -0.04279788,
    #                -0.04734922, -0.033705, -0.02380217, -0.04480772, -0.03756881]
    # upper_guess = [0.17240303, 0.26668075, 0.21499604, 0.26299318, 0.22545807,
    #                0.24759903, 0.31644402, 0.2964658, 0.15360716, 0.31317824,
    #                0.27760982, 0.23009955, 0.24045039, 0.37542525, 0.21361931,
    #                0.18678503, 0.23466624, 0.20630533, 0.16191541, 0.20453953,
    #                0.14370825, 0.13428077, 0.15387739, 0.13767285, 0.15173257,
    #                0.14042002, 0.11336701, 0.35640688, 0.10953915, 0.08167446]

    ### Packing/Unpacking functions
    n_lower = len(lower_guess)
    n_upper = len(upper_guess)
    pack = lambda lower, upper: np.concatenate((lower, upper))
    unpack = lambda pack: (pack[:n_lower], pack[n_lower:])


    def make_airfoil(x):
        """
        A function that constructs an airfoil from a packed design vector.
        :param x:
        :return:
        """
        lower, upper = unpack(x)
        return Airfoil(
            name="Optimization Airfoil",
            coordinates=get_kulfan_coordinates(
                lower_weights=lower,
                upper_weights=upper,
                enforce_continuous_LE_radius=enforce_continuous_LE_radius,
                TE_thickness=TE_thickness,
                n_points_per_side=80
            )
        )


    ### Initial guess construction
    x0 = pack(lower_guess, upper_guess)
    initial_airfoil = make_airfoil(x0)

    ### Initialize plotting
    fig = plt.figure(figsize=(15, 2.5))
    ax = fig.add_subplot(111)
    trace_initial, = ax.plot(
        initial_airfoil.coordinates[:, 0],
        initial_airfoil.coordinates[:, 1],
        ':r',
        label="Initial Airfoil"
    )
    trace_current, = ax.plot(
        initial_airfoil.coordinates[:, 0],
        initial_airfoil.coordinates[:, 1],
        "-b",
        label="Current Airfoil"
    )
    plt.axis("equal")
    plt.xlabel(r"$x/c$")
    plt.ylabel(r"$y/c$")
    plt.title("Airfoil Optimization")
    plt.legend()


    def draw(
            airfoil  # type: Airfoil
    ):
        """
        Updates the "current airfoil" line on the plot with the given airfoil.
        :param airfoil:
        :return:
        """
        trace_current.set_xdata(airfoil.coordinates[:, 0])
        trace_current.set_ydata(airfoil.coordinates[:, 1])
        plt.draw()
        plt.pause(0.001)


    ### Utilities for tracking the design vector and objective throughout the optimization run
    iteration = 0
    xs = []
    fs = []


    def augmented_objective(x):
        """
        Objective function with constraints added via a multiplicative external penalty method
        :param x: Packed design vector
        :return: Value of the augmented objective
        """
        airfoil = make_airfoil(x)
        xfoil = airfoil.xfoil_cseq(
            cl_start=Cl_start,
            cl_step=0.02,
            cl_end=Cl_end,
            Re=Re_des,
            verbose=False,
            max_iter=40,
            repanel=False
        )
        if np.isnan(xfoil["Cd"]).any():
            return np.inf

        objective = np.sqrt(np.mean(xfoil["Cd"] ** 2))  # RMS

        penalty = 0
        penalty += np.sum(np.minimum(0, (xfoil["Cm"] - Cm_min) / 0.01) ** 2)  # Cm constraint
        penalty += np.minimum(0, (airfoil.TE_angle() - 5) / 1) ** 2  # TE angle constraint
        penalty += np.minimum(0, (airfoil.local_thickness(0.90) - 0.015) / 0.005) ** 2  # Spar thickness constraint
        penalty += np.minimum(0, (airfoil.local_thickness(0.30) - 0.12) / 0.005) ** 2  # Spar thickness constraint

        xs.append(x)
        fs.append(objective)

        return objective * (1 + penalty)


    def callback(x):
        global iteration
        iteration += 1
        print(
            f"Iteration {iteration}: Cd = {fs[-1]:.6f}"
        )
        if iteration % 1 == 0:
            airfoil = make_airfoil(x)
            draw(airfoil)
            ax.set_title(f"Airfoil Optimization: Iteration {iteration}")
            airfoil.write_dat("optimized_airfoil.dat")


    draw(initial_airfoil)

    initial_simplex = (
            (0.5 + 1 * np.random.random((len(x0) + 1, len(x0))))
            * x0
    )
    initial_simplex[0, :] = x0  # Include x0 in the simplex
    print("Initializing simplex (give this a few minutes)...")
    res = optimize.minimize(
        fun=augmented_objective,
        x0=pack(lower_guess, upper_guess),
        method="Nelder-Mead",
        callback=callback,
        options={
            'maxiter'        : 10 ** 6,
            'initial_simplex': initial_simplex,
            'xatol'          : 1e-8,
            'fatol'          : 1e-6,
            'adaptive'       : False,
        }
    )

    final_airfoil = make_airfoil(res.x)



from aerosandbox import AeroSandboxObject
from aerosandbox.geometry.common import *
from typing import List, Dict, Any, Union, Optional, Tuple
import aerosandbox.geometry.mesh_utilities as mesh_utils
from aerosandbox.geometry.wing import Wing
from aerosandbox.geometry.fuselage import Fuselage
from aerosandbox.geometry.propulsor import Propulsor
from aerosandbox.weights.mass_properties import MassProperties
import copy
from pathlib import Path


class Airplane(AeroSandboxObject):
    """
    Definition for an airplane.

    Anatomy of an Airplane:

        An Airplane consists chiefly of a collection of wings and fuselages. These can be accessed with
        `Airplane.wings` and `Airplane.fuselages`, which gives a list of those respective components. Each wing is a
        Wing object, and each fuselage is a Fuselage object.

    """

    def __init__(self,
                 name: Optional[str] = None,
                 xyz_ref: Union[np.ndarray, List] = None,
                 wings: Optional[List[Wing]] = None,
                 fuselages: Optional[List[Fuselage]] = None,
                 propulsors: Optional[List[Propulsor]] = None,
                 s_ref: Optional[float] = None,
                 c_ref: Optional[float] = None,
                 b_ref: Optional[float] = None,
                 analysis_specific_options: Optional[Dict[type, Dict[str, Any]]] = None
                 ):
        """
        Defines a new airplane.

        Args:

            name: Name of the airplane [optional]. It can help when debugging to give the airplane a sensible name.

            xyz_ref: An array-like that gives the x-, y-, and z- reference point of the airplane, used when computing
            moments and stability derivatives. Generally, this should be the center of gravity.

                # In a future version, this will be deprecated and replaced with asb.MassProperties.

            wings: A list of Wing objects that are a part of the airplane.

            fuselages: A list of Fuselage objects that are a part of the airplane.

            propulsors: A list of Propulsor objects that are a part of the airplane.

            s_ref: Reference area. If undefined, it's set from the area of the first Wing object. # Note: will be deprecated

            c_ref: Reference chord. If undefined, it's set from the mean aerodynamic chord of the first Wing object. # Note: will be deprecated

            b_ref: Reference span. If undefined, it's set from the span of the first Wing object. # Note: will be deprecated

            analysis_specific_options: Analysis-specific options are additional constants or modeling assumptions
            that should be passed on to specific analyses and associated with this specific geometry object.

                This should be a dictionary where:

                    * Keys are specific analysis types (typically a subclass of asb.ExplicitAnalysis or
                    asb.ImplicitAnalysis), but if you decide to write your own analysis and want to make this key
                    something else (like a string), that's totally fine - it's just a unique identifier for the
                    specific analysis you're running.

                    * Values are a dictionary of key:value pairs, where:

                        * Keys are strings.

                        * Values are some value you want to assign.

                This is more easily demonstrated / understood with an example:

                >>> analysis_specific_options = {
                >>>     asb.AeroBuildup: dict(
                >>>         include_wave_drag=True,
                >>>     )
                >>> }

        """
        ### Set defaults
        if name is None:
            name = "Untitled"
        if xyz_ref is None:
            xyz_ref = np.array([0., 0., 0.])
        if wings is None:
            wings: List[Wing] = []
        if fuselages is None:
            fuselages: List[Fuselage] = []
        if propulsors is None:
            propulsors: List[Propulsor] = []
        if analysis_specific_options is None:
            analysis_specific_options = {}

        ### Initialize
        self.name = name
        self.xyz_ref = np.array(xyz_ref)
        self.wings = wings
        self.fuselages = fuselages
        self.propulsors = propulsors
        self.analysis_specific_options = analysis_specific_options

        ### Assign reference values
        try:
            main_wing = self.wings[0]
        except IndexError:
            main_wing = None

        try:
            main_fuse = self.fuselages[0]
        except IndexError:
            main_fuse = None

        if s_ref is not None:
            self.s_ref = s_ref
        else:
            if main_wing is not None:
                self.s_ref = main_wing.area()
            else:
                if main_fuse is not None:
                    self.s_ref = main_fuse.area_projected()
                else:
                    raise ValueError(
                        "`s_ref` was not provided, and a value cannot be inferred automatically from wings or fuselages.\n"
                        "You must set this manually when instantiating your asb.Airplane object.")

        if c_ref is not None:
            self.c_ref = c_ref
        else:
            if main_wing is not None:
                self.c_ref = main_wing.mean_aerodynamic_chord()
            else:
                if main_fuse is not None:
                    self.c_ref = main_fuse.length()
                else:
                    raise ValueError(
                        "`c_ref` was not provided, and a value cannot be inferred automatically from wings or fuselages.\n"
                        "You must set this manually when instantiating your asb.Airplane object."
                    )

        if b_ref is not None:
            self.b_ref = b_ref
        else:
            if main_wing is not None:
                self.b_ref = main_wing.span(include_centerline_distance=True)
            else:
                if main_fuse is not None:
                    self.b_ref = main_fuse.area_projected() / main_fuse.length()
                else:
                    raise ValueError(
                        "`b_ref` was not provided, and a value cannot be inferred automatically from wings or fuselages.\n"
                        "You must set this manually when instantiating your asb.Airplane object."
                    )

    def __repr__(self):
        n_wings = len(self.wings)
        n_fuselages = len(self.fuselages)
        return f"Airplane '{self.name}' " \
               f"({n_wings} {'wing' if n_wings == 1 else 'wings'}, " \
               f"{n_fuselages} {'fuselage' if n_fuselages == 1 else 'fuselages'})"

    # TODO def add_wing(wing: 'Wing') -> None

    def mesh_body(self,
                  method="quad",
                  thin_wings=False,
                  stack_meshes=True,
                  ):
        """
        Returns a surface mesh of the Airplane, in (points, faces) format. For reference on this format,
        see the documentation in `aerosandbox.geometry.mesh_utilities`.

        Args:

            method:

            thin_wings: Controls whether wings should be meshed as thin surfaces, rather than full 3D bodies.

            stack_meshes: Controls whether the meshes should be merged into a single mesh or not.

                * If True, returns a (points, faces) tuple in standard mesh format.

                * If False, returns a list of (points, faces) tuples in standard mesh format.

        Returns:

        """
        if thin_wings:
            wing_meshes = [
                wing.mesh_thin_surface(
                    method=method,
                )
                for wing in self.wings
            ]
        else:
            wing_meshes = [
                wing.mesh_body(
                    method=method,
                )
                for wing in self.wings
            ]

        fuse_meshes = [
            fuse.mesh_body(
                method=method
            )
            for fuse in self.fuselages
        ]

        meshes = wing_meshes + fuse_meshes

        if stack_meshes:
            points, faces = mesh_utils.stack_meshes(*meshes)
            return points, faces
        else:
            return meshes

    def draw(self,
             backend: str = "pyvista",
             thin_wings: bool = False,
             ax=None,
             use_preset_view_angle: str = None,
             set_background_pane_color: Union[str, Tuple[float, float, float]] = None,
             set_background_pane_alpha: float = None,
             set_lims: bool = True,
             set_equal: bool = True,
             set_axis_visibility: bool = None,
             show: bool = True,
             show_kwargs: Dict = None,
             ):
        """
        Produces an interactive 3D visualization of the airplane.

        Args:

            backend: The visualization backend to use. Options are:

                * "matplotlib" for a Matplotlib backend
                * "pyvista" for a PyVista backend
                * "plotly" for a Plot.ly backend
                * "trimesh" for a trimesh backend

            thin_wings: A boolean that determines whether to draw the full airplane (i.e. thickened, 3D bodies), or to use a
            thin-surface representation for any Wing objects.

            show: A boolean that determines whether to display the object after plotting it. If False, the object is
            returned but not displayed. If True, the object is displayed and returned.

        Returns: The plotted object, in its associated backend format. Also displays the object if `show` is True.

        """
        if show_kwargs is None:
            show_kwargs = {}

        points, faces = self.mesh_body(method="quad", thin_wings=thin_wings)

        if backend == "matplotlib":
            import matplotlib.pyplot as plt
            import aerosandbox.tools.pretty_plots as p
            from mpl_toolkits.mplot3d.art3d import Poly3DCollection

            if ax is None:
                _, ax = p.figure3d(figsize=(8, 8), computed_zorder=False)

            else:
                if not p.ax_is_3d(ax):
                    raise ValueError("`ax` must be a 3D axis.")

            plt.sca(ax)

            ### Set the view angle
            if use_preset_view_angle is not None:
                p.set_preset_3d_view_angle(use_preset_view_angle)

            ### Set the background pane color
            if set_background_pane_color is not None:
                ax.xaxis.pane.set_facecolor(set_background_pane_color)
                ax.yaxis.pane.set_facecolor(set_background_pane_color)
                ax.zaxis.pane.set_facecolor(set_background_pane_color)

            ### Set the background pane alpha
            if set_background_pane_alpha is not None:
                ax.xaxis.pane.set_alpha(set_background_pane_alpha)
                ax.yaxis.pane.set_alpha(set_background_pane_alpha)
                ax.zaxis.pane.set_alpha(set_background_pane_alpha)

            ax.add_collection(
                Poly3DCollection(
                    points[faces], facecolors='lightgray', edgecolors=(0, 0, 0, 0.1),
                    linewidths=0.5, alpha=0.8, shade=True,
                ),
            )

            for prop in self.propulsors:

                ### Disk
                if prop.length == 0:
                    ax.add_collection(
                        Poly3DCollection(
                            np.stack([np.stack(
                                prop.get_disk_3D_coordinates(),
                                axis=1
                            )], axis=0),
                            facecolors='darkgray', edgecolors=(0, 0, 0, 0.2),
                            linewidths=0.5, alpha=0.35, shade=True, zorder=4,
                        )
                    )

            if set_lims:
                ax.set_xlim(points[:, 0].min(), points[:, 0].max())
                ax.set_ylim(points[:, 1].min(), points[:, 1].max())
                ax.set_zlim(points[:, 2].min(), points[:, 2].max())

            if set_equal:
                p.equal()

            if set_axis_visibility is not None:
                if set_axis_visibility:
                    ax.set_axis_on()
                else:
                    ax.set_axis_off()

            if show:
                p.show_plot()

        elif backend == "plotly":

            from aerosandbox.visualization.plotly_Figure3D import Figure3D
            fig = Figure3D()
            for f in faces:
                fig.add_quad((
                    points[f[0]],
                    points[f[1]],
                    points[f[2]],
                    points[f[3]],
                ), outline=True)
                show_kwargs = {
                    "show": show,
                    **show_kwargs
                }
            return fig.draw(**show_kwargs)

        elif backend == "pyvista":

            import pyvista as pv
            fig = pv.PolyData(
                *mesh_utils.convert_mesh_to_polydata_format(points, faces)
            )
            show_kwargs = {
                "show_edges": True,
                "show_grid" : True,
                **show_kwargs,
            }
            if show:
                fig.plot(**show_kwargs)
            return fig

        elif backend == "trimesh":

            import trimesh as tri
            fig = tri.Trimesh(points, faces)
            if show:
                fig.show(**show_kwargs)
            return fig
        else:
            raise ValueError("Bad value of `backend`!")

    def draw_wireframe(self,
                       ax=None,
                       color="k",
                       thin_linewidth=0.2,
                       thick_linewidth=0.5,
                       fuselage_longeron_theta=None,
                       use_preset_view_angle: str = None,
                       set_background_pane_color: Union[str, Tuple[float, float, float]] = None,
                       set_background_pane_alpha: float = None,
                       set_lims: bool = True,
                       set_equal: bool = True,
                       set_axis_visibility: bool = None,
                       show: bool = True,
                       ) -> "matplotlib.axes.Axes":
        """
        Draws a wireframe of the airplane on a Matplotlib 3D axis.

        Args:

            ax: The axis to draw on. Must be a 3D axis. If None, creates a new axis.

            color: The color of the wireframe.

            thin_linewidth: The linewidth of the thin lines.
        """
        ### Set defaults
        if fuselage_longeron_theta is None:
            fuselage_longeron_theta = np.linspace(0, 2 * np.pi, 8 + 1)[:-1]

        import matplotlib.pyplot as plt
        import aerosandbox.tools.pretty_plots as p

        if ax is None:
            _, ax = p.figure3d(figsize=(8, 8), computed_zorder=False)

        else:
            if not p.ax_is_3d(ax):
                raise ValueError("`ax` must be a 3D axis.")

        plt.sca(ax)

        ### Set the view angle
        if use_preset_view_angle is not None:
            p.set_preset_3d_view_angle(use_preset_view_angle)

        ### Set the background pane color
        if set_background_pane_color is not None:
            ax.xaxis.pane.set_facecolor(set_background_pane_color)
            ax.yaxis.pane.set_facecolor(set_background_pane_color)
            ax.zaxis.pane.set_facecolor(set_background_pane_color)

        ### Set the background pane alpha
        if set_background_pane_alpha is not None:
            ax.xaxis.pane.set_alpha(set_background_pane_alpha)
            ax.yaxis.pane.set_alpha(set_background_pane_alpha)
            ax.zaxis.pane.set_alpha(set_background_pane_alpha)

        def plot_line(
                xyz: np.ndarray,
                symmetric: bool = False,
                color=color,
                linewidth=0.4,
                **kwargs
        ):
            if symmetric:
                xyz = np.concatenate([
                    xyz,
                    np.array([[np.nan] * 3]),
                    xyz * np.array([[1, -1, 1]])
                ], axis=0)

            ax.plot(
                xyz[:, 0],
                xyz[:, 1],
                xyz[:, 2],
                color=color,
                linewidth=linewidth,
                **kwargs
            )

        def reshape(x):
            return np.reshape(np.array(x), (1, 3))

        ##### Wings
        for wing in self.wings:
            try:
                if wing.color is not None:
                    color_to_use = wing.color
                else:
                    color_to_use = color
            except AttributeError:
                color_to_use = color

            ### LE and TE lines
            for xy in [
                (0, 0),  # Leading Edge
                (1, 0),  # Trailing Edge
            ]:

                plot_line(
                    np.stack(wing.mesh_line(x_nondim=xy[0], z_nondim=xy[1]), axis=0),
                    symmetric=wing.symmetric,
                    linewidth=thick_linewidth,
                    color=color_to_use
                )

            ### Top and Bottom lines
            x = 0.4
            afs = [xsec.airfoil for xsec in wing.xsecs]
            thicknesses = np.array([af.local_thickness(x_over_c=x) for af in afs])

            plot_line(
                np.stack(wing.mesh_line(x_nondim=x, z_nondim=thicknesses / 2, add_camber=True), axis=0),
                symmetric=wing.symmetric,
                linewidth=thin_linewidth,
                color=color_to_use
            )
            plot_line(
                np.stack(wing.mesh_line(x_nondim=x, z_nondim=-thicknesses / 2, add_camber=True), axis=0),
                symmetric=wing.symmetric,
                linewidth=thin_linewidth,
                color=color_to_use
            )

            ### Airfoils
            for i, xsec in enumerate(wing.xsecs):
                xg_local, yg_local, zg_local = wing._compute_frame_of_WingXSec(i)
                xg_local = reshape(xg_local)
                yg_local = reshape(yg_local)
                zg_local = reshape(zg_local)
                origin = reshape(xsec.xyz_le)
                scale = xsec.chord

                line_upper = origin + (
                        xsec.airfoil.upper_coordinates()[:, 0].reshape((-1, 1)) * scale * xg_local +
                        xsec.airfoil.upper_coordinates()[:, 1].reshape((-1, 1)) * scale * zg_local
                )
                line_lower = origin + (
                        xsec.airfoil.lower_coordinates()[:, 0].reshape((-1, 1)) * scale * xg_local +
                        xsec.airfoil.lower_coordinates()[:, 1].reshape((-1, 1)) * scale * zg_local
                )

                for line in [line_upper, line_lower]:
                    plot_line(
                        line,
                        symmetric=wing.symmetric,
                        linewidth=thick_linewidth if i == 0 or i == len(wing.xsecs) - 1 else thin_linewidth,
                        color=color_to_use
                    )

        ##### Fuselages
        for fuse in self.fuselages:
            try:
                if fuse.color is not None:
                    color_to_use = fuse.color
                else:
                    color_to_use = color
            except AttributeError:
                color_to_use = color

            ### Bulkheads
            perimeters_xyz = [
                xsec.get_3D_coordinates(theta=np.linspace(0, 2 * np.pi, 121))
                for xsec in fuse.xsecs
            ]
            for i, perim in enumerate(perimeters_xyz):
                plot_line(
                    np.stack(perim, axis=1),
                    linewidth=thick_linewidth if i == 0 or i == len(fuse.xsecs) - 1 else thin_linewidth,
                    color=color_to_use
                )

            ### Centerline
            plot_line(
                np.stack(
                    fuse.mesh_line(y_nondim=0, z_nondim=0),
                    axis=0,
                ),
                linewidth=thin_linewidth,
                color=color_to_use
            )

            ### Longerons
            for theta in fuselage_longeron_theta:
                plot_line(
                    np.stack([
                        np.array(xsec.get_3D_coordinates(theta=theta))
                        for xsec in fuse.xsecs
                    ], axis=0),
                    linewidth=thick_linewidth,
                    color=color_to_use
                )

        ##### Propulsors
        for prop in self.propulsors:
            try:
                if prop.color is not None:
                    color_to_use = prop.color
                else:
                    color_to_use = color
            except AttributeError:
                color_to_use = color

            ### Disk
            if prop.length == 0:
                plot_line(
                    np.stack(
                        prop.get_disk_3D_coordinates(),
                        axis=1
                    ),
                    color=color_to_use
                )

        if set_lims:
            points, _ = self.mesh_body()
            ax.set_xlim(points[:, 0].min(), points[:, 0].max())
            ax.set_ylim(points[:, 1].min(), points[:, 1].max())
            ax.set_zlim(points[:, 2].min(), points[:, 2].max())

        if set_equal:
            p.equal()

        if set_axis_visibility is not None:
            if set_axis_visibility:
                ax.set_axis_on()
            else:
                ax.set_axis_off()

        if show:
            p.show_plot()

        return ax

    def draw_three_view(self,
                        axs=None,
                        style: str = "shaded",
                        show: bool = True,
                        ) -> np.ndarray:
        """
        Draws a standard 4-panel three-view diagram of the airplane using Matplotlib backend. Creates a new figure.

        Args:

            axs: A 2D numpy array of Matplotlib axes objects, with shape at least (2, 2). Each subplot must be 3D. If
            None, creates a new figure with the required axes.

            style: Determines what drawing style to use for the three-view. A string, one of:

                * "shaded"
                * "wireframe"

            show: A Boolean of whether to show the figure after creating it, or to hold it so that the user can
            modify the figure further before showing.

        Returns: A 2D NumPy array of Matplotlib axes objects, with shape (2, 2). The axes are arranged as follows:

                * axs[0, 0]: Top view
                * axs[0, 1]: Front view
                * axs[1, 0]: Side view
                * axs[1, 1]: Isometric view
        """
        import matplotlib.pyplot as plt
        import aerosandbox.tools.pretty_plots as p

        preset_view_angles = np.array([
            ["XZ", "-YZ"],
            ["XY", "left_isometric"]
        ], dtype="O")

        if axs is None:
            fig, axs = p.figure3d(
                nrows=preset_view_angles.shape[0],
                ncols=preset_view_angles.shape[1],
                figsize=(8, 8),
                computed_zorder=False,
            )
        else:
            if not len(axs.shape) == 2:
                raise ValueError(f"`axs` must be a 2D array of axes; instead, it is: {axs}.")
            if not axs.shape[0] >= preset_view_angles.shape[0]:
                raise ValueError(
                    f"`axs` must have at least as many rows as preset_view_angles ({preset_view_angles.shape[0]}).")
            if not axs.shape[1] >= preset_view_angles.shape[1]:
                raise ValueError(
                    f"`axs` must have at least as many columns as preset_view_angles ({preset_view_angles.shape[1]}).")

        for i in range(axs.shape[0]):
            for j in range(axs.shape[1]):
                ax = axs[i, j]
                preset_view = preset_view_angles[i, j]

                if style == "shaded":
                    self.draw(
                        backend="matplotlib",
                        ax=ax,
                        set_axis_visibility=False if 'isometric' in preset_view else None,
                        show=False
                    )
                elif style == "wireframe":
                    if preset_view == "XZ":
                        fuselage_longeron_theta = [np.pi / 2, 3 * np.pi / 2]
                    elif preset_view == "XY":
                        fuselage_longeron_theta = [0, np.pi]
                    else:
                        fuselage_longeron_theta = None

                    self.draw_wireframe(
                        ax=ax,
                        set_axis_visibility=False if 'isometric' in preset_view else None,
                        fuselage_longeron_theta=fuselage_longeron_theta,
                        show=False
                    )

                p.set_preset_3d_view_angle(preset_view)

                if preset_view == 'XY' or preset_view == '-XY':
                    ax.set_zticks([])
                if preset_view == 'XZ' or preset_view == '-XZ':
                    ax.set_yticks([])
                if preset_view == 'YZ' or preset_view == '-YZ':
                    ax.set_xticks([])

        axs[1, 0].set_xlabel("$x_g$ [m]")
        axs[1, 0].set_ylabel("$y_g$ [m]")
        axs[0, 0].set_zlabel("$z_g$ [m]")
        axs[0, 0].set_xticklabels([])
        axs[0, 1].set_yticklabels([])
        axs[0, 1].set_zticklabels([])

        plt.subplots_adjust(
            left=-0.08,
            right=1.08,
            bottom=-0.08,
            top=1.08,
            wspace=-0.38,
            hspace=-0.38,
        )

        if show:
            p.show_plot(
                tight_layout=False,
            )

        return axs

    def is_entirely_symmetric(self):
        """
        Returns a boolean describing whether the airplane is geometrically entirely symmetric across the XZ-plane.
        :return: [boolean]
        """
        for wing in self.wings:
            if not wing.is_entirely_symmetric():
                return False

        # TODO add in logic for fuselages

        return True

    def aerodynamic_center(self, chord_fraction: float = 0.25):
        """
        Computes the approximate location of the aerodynamic center of the wing.
        Uses the generalized methodology described here:
            https://core.ac.uk/download/pdf/79175663.pdf

        Args:

            chord_fraction: The position of the aerodynamic center along the MAC, as a fraction of MAC length.
            Typically, this value (denoted `h_0` in the literature) is 0.25 for a subsonic wing. However,
            wing-fuselage interactions can cause a forward shift to a value more like 0.1 or less. Citing Cook,
            Michael V., "Flight Dynamics Principles", 3rd Ed., Sect. 3.5.3 "Controls-fixed static stability". PDF:
            https://www.sciencedirect.com/science/article/pii/B9780080982427000031

        Returns: The (x, y, z) coordinates of the aerodynamic center of the airplane.
        """
        wing_areas = [wing.area(type="projected") for wing in self.wings]
        ACs = [wing.aerodynamic_center() for wing in self.wings]

        wing_AC_area_products = [
            AC * area
            for AC, area in zip(
                ACs,
                wing_areas
            )
        ]

        aerodynamic_center = sum(wing_AC_area_products) / sum(wing_areas)

        return aerodynamic_center

    def with_control_deflections(self,
                                 control_surface_deflection_mappings: Dict[str, float]
                                 ) -> "Airplane":
        """
        Returns a copy of the airplane with the specified control surface deflections applied.

        Args:
            control_surface_deflection_mappings: A dictionary mapping control surface names to deflections.

                * Keys: Control surface names.

                * Values: Deflections, in degrees. Downwards-positive, following typical convention.

        Returns: A copy of the airplane with the specified control surface deflections applied.

        """
        deflected_airplane = copy.deepcopy(self)

        for name, deflection in control_surface_deflection_mappings.items():

            for wi, wing in enumerate(deflected_airplane.wings):

                for xi, xsec in enumerate(wing.xsecs):

                    for csi, surf in enumerate(xsec.control_surfaces):

                        if surf.name == name:

                            surf.deflection = deflection

        return deflected_airplane

    def generate_cadquery_geometry(self,
                                   minimum_airfoil_TE_thickness: float = 0.001,
                                   fuselage_tol: float = 1e-4,
                                   ) -> "Workplane":
        """
        Uses the CADQuery library (OpenCASCADE backend) to generate a 3D CAD model of the airplane.

        Args:

            minimum_airfoil_TE_thickness: The minimum thickness of the trailing edge of the airfoils, as a fraction
            of each airfoil's chord. This will be enforced by thickening the trailing edge of the airfoils if
            necessary. This is useful for avoiding numerical issues in CAD software that can arise from extremely
            thin (i.e., <1e-6 meters) trailing edges.

            tol: The geometric tolerance (meters) to use when generating the CAD geometry. This is passed directly to the CADQuery

        Returns: A CADQuery Workplane object containing the CAD geometry of the airplane.

        """
        try:
            import cadquery as cq
        except ModuleNotFoundError:
            raise ModuleNotFoundError(
                "The `cadquery` library is required to use this function. Please install it with `pip install cadquery`.")

        solids = []

        for wing in self.wings:

            xsec_wires = []

            for i, xsec in enumerate(wing.xsecs):
                csys = wing._compute_frame_of_WingXSec(i)

                af = xsec.airfoil
                if af.TE_thickness() < minimum_airfoil_TE_thickness:
                    af = af.set_TE_thickness(
                        thickness=minimum_airfoil_TE_thickness
                    )

                LE_index = af.LE_index()

                xsec_wires.append(
                    cq.Workplane(
                        inPlane=cq.Plane(
                            origin=tuple(xsec.xyz_le),
                            xDir=tuple(csys[0]),
                            normal=tuple(-csys[1])
                        )
                    ).spline(
                        listOfXYTuple=[
                            tuple(xy * xsec.chord)
                            for xy in af.coordinates[:LE_index, :]
                        ]
                    ).spline(
                        listOfXYTuple=[
                            tuple(xy * xsec.chord)
                            for xy in af.coordinates[LE_index:, :]
                        ]
                    ).close()
                )

            wire_collection = xsec_wires[0]
            for s in xsec_wires[1:]:
                wire_collection.ctx.pendingWires.extend(s.ctx.pendingWires)

            loft = wire_collection.loft(ruled=True, clean=False)

            solids.append(loft)

            if wing.symmetric:
                loft = loft.mirror(
                    mirrorPlane='XZ',
                    union=False
                )

                solids.append(loft)

        for fuse in self.fuselages:

            xsec_wires = []

            for i, xsec in enumerate(fuse.xsecs):

                if xsec.height < fuselage_tol or xsec.width < fuselage_tol:  # If the xsec is so small as to effectively be a point
                    xsec = copy.deepcopy(xsec)  # Modify the xsec to be big enough to not error out.
                    xsec.height = np.maximum(xsec.height, fuselage_tol)
                    xsec.width = np.maximum(xsec.width, fuselage_tol)

                xsec_wires.append(
                    cq.Workplane(
                        inPlane=cq.Plane(
                            origin=tuple(xsec.xyz_c),
                            xDir=(0, 1, 0),
                            normal=(-1, 0, 0)
                        )
                    ).spline(
                        listOfXYTuple=[
                            (y - xsec.xyz_c[1], z - xsec.xyz_c[2])
                            for x, y, z in zip(*xsec.get_3D_coordinates(
                                theta=np.linspace(
                                    np.pi / 2, np.pi / 2 + 2 * np.pi,
                                    181
                                )
                            ))
                        ]
                    ).close()
                )

            wire_collection = xsec_wires[0]
            for s in xsec_wires[1:]:
                wire_collection.ctx.pendingWires.extend(s.ctx.pendingWires)

            loft = wire_collection.loft(ruled=True, clean=False)

            solids.append(loft)

        solid = solids[0]
        for s in solids[1:]:
            solid.add(s)

        return solid.clean()

    def export_cadquery_geometry(self,
                                 filename: Union[Path, str],
                                 minimum_airfoil_TE_thickness: float = 0.001
                                 ) -> None:
        """
        Exports the airplane geometry to a STEP file.

        Args:
            filename: The filename to export to. Should include the ".step" extension.

            minimum_airfoil_TE_thickness: The minimum thickness of the trailing edge of the airfoils, as a fraction
            of each airfoil's chord. This will be enforced by thickening the trailing edge of the airfoils if
            necessary. This is useful for avoiding numerical issues in CAD software that can arise from extremely
            thin (i.e., <1e-6 meters) trailing edges.

        Returns: None, but exports the airplane geometry to a STEP file.
        """
        solid = self.generate_cadquery_geometry(
            minimum_airfoil_TE_thickness=minimum_airfoil_TE_thickness,
        )

        solid.objects = [
            o.scale(1000)
            for o in solid.objects
        ]

        from cadquery import exporters
        exporters.export(
            solid,
            fname=filename
        )

    def export_AVL(self,
                   filename,
                   include_fuselages: bool = True
                   ):
        # TODO include option for mass file export as well
        # Use MassProperties.export_AVL_mass...

        from aerosandbox.aerodynamics.aero_3D.avl import AVL
        avl = AVL(
            airplane=self,
            op_point=None,
            xyz_ref=self.xyz_ref
        )
        avl.write_avl(filepath=filename)

    def export_XFLR(self, *args, **kwargs) -> str:
        import warnings

        warnings.warn(
            "`Airplane.export_XFLR()` has been renamed to `Airplane.export_XFLR5_xml()`, to clarify\n"
            "that it exports to XFLR5's XML format, not to a XFL file.\n"
            "\n"
            "Please update your code to use `Airplane.export_XFLR5_xml()` instead.\n"
            "\n"
            "This function will be removed in a future version of AeroSandbox.",
            PendingDeprecationWarning
        )
        return self.export_XFLR5_xml(*args, **kwargs)

    def export_XFLR5_xml(self,
                         filename: Union[Path, str],
                         mass_props: MassProperties = None,
                         include_fuselages: bool = False,
                         mainwing: Wing = None,
                         elevator: Wing = None,
                         fin: Wing = None,
                         ) -> str:
        """
        Exports the airplane geometry to an XFLR5 `.xml` file. To import the `.xml` file into XFLR5, go to File ->
        Import -> Import from XML.

        Args:
            filename: The filename to export to. Should include the ".xml" extension.

            mass_props: The MassProperties object to use when exporting the airplane. If not specified, will default to
                a 1 kg point mass at the origin.

                - Note: XFLR5 does not natively support user-defined inertia tensors, so we have to synthesize an equivalent
                set of point masses to represent the inertia tensor.

            include_fuselages: Whether to include fuselages in the export.

            mainwing: The main wing of the airplane. If not specified, will default to the first wing in the airplane.

            elevator: The elevator of the airplane. If not specified, will default to the second wing in the airplane.

            fin: The fin of the airplane. If not specified, will default to the third wing in the airplane.

        Returns: None, but exports the airplane geometry to an XFLR5 `.xml` file.

            To import the `.xml` file into XFLR5, go to File -> Import -> Import from XML.
        """
        ### Handle default arguments
        if mass_props is None:
            mass_props = MassProperties(
                mass=1,
                x_cg=0,
                y_cg=0,
                z_cg=0,
            )

        ### Identify which wings are the main wing, elevator, and fin.
        wings_specified = [
            mainwing is not None,
            elevator is not None,
            fin is not None,
        ]
        if all(wings_specified):
            pass
        elif any(wings_specified):
            raise ValueError(
                "If any wings are specified (`mainwing`, `elevator`, `fin`), then all wings must be specified.")
        else:
            n_wings = len(self.wings)

            if n_wings == 0:
                pass
            else:
                import warnings
                warnings.warn(
                    "No wings were specified (`mainwing`, `elevator`, `fin`). Automatically assigning the first wing "
                    "to `mainwing`, the second wing to `elevator`, and the third wing to `fin`. If this is not "
                    "correct, manually specify these with (`mainwing`, `elevator`, and `fin`) arguments."
                )

                if n_wings == 1:
                    mainwing = self.wings[0]
                elif n_wings == 2:
                    mainwing = self.wings[0]
                    elevator = self.wings[1]
                elif n_wings == 3:
                    mainwing = self.wings[0]
                    elevator = self.wings[1]
                    fin = self.wings[2]
                else:
                    raise ValueError(
                        "Could not automatically parse which wings should be assigned to which XFLR5 lifting surfaces, "
                        "since there are too many. Manually assign these with (`mainwing`, `elevator`, and `fin`) "
                        "arguments."
                    )

        ### Determine where point masses should be in order to yield the specified mass properties.
        point_masses = mass_props.generate_possible_set_of_point_masses()

        ### Handle the fuselage
        if include_fuselages:
            raise NotImplementedError(
                "Fuselage export to XFLR5 is not yet implemented."
            )

        ### Write the XML file.
        import xml.etree.ElementTree as ET

        base_xml = f"""\
<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE explane>
<explane version="1.0">
    <Units>
        <length_unit_to_meter>1</length_unit_to_meter>
        <mass_unit_to_kg>1</mass_unit_to_kg>
    </Units>
    <Plane>
        <Name>{self.name}</Name>
        <Description></Description>
        <Inertia>
        </Inertia>
        <has_body>false</has_body>
    </Plane>
</explane>
"""

        root = ET.fromstring(base_xml)
        plane = root.find("Plane")

        ### Add point masses
        inertia = plane.find("Inertia")
        for i, point_mass in enumerate(point_masses):
            point_mass_xml = ET.SubElement(inertia, "Point_Mass")

            for k, v in {
                "Tag"        : f"pm{i}",
                "Mass"       : point_mass.mass,
                "coordinates": ",".join([str(x) for x in point_mass.xyz_cg]),
            }.items():
                subelement = ET.SubElement(point_mass_xml, k)
                subelement.text = str(v)

        ### Add the wings
        if mainwing is not None:
            wing = mainwing
            wingxml = ET.SubElement(plane, "wing")

            xyz_le_root = wing._compute_xyz_of_WingXSec(index=0, x_nondim=0, z_nondim=0)

            for k, v in {
                "Name"      : wing.name,
                "Type"      : "MAINWING",
                "Position"  : ",".join([str(x) for x in xyz_le_root]),
                "Tilt_angle": 0.,
                "Symetric"  : wing.symmetric,  # This tag is a typo in XFLR...
                "isFin"     : "false",
                "isSymFin"  : "false",
            }.items():
                subelement = ET.SubElement(wingxml, k)
                subelement.text = str(v)

            sections = ET.SubElement(wingxml, "Sections")

            xyz_le_sects_rel = [
                wing._compute_xyz_of_WingXSec(index=i, x_nondim=0, z_nondim=0) - xyz_le_root
                for i in range(len(wing.xsecs))
            ]

            for i, xsec in enumerate(wing.xsecs):

                sect = ET.SubElement(sections, "Section")

                if i == len(wing.xsecs) - 1:
                    dihedral = 0
                else:
                    dihedral = np.arctan2d(
                        xyz_le_sects_rel[i + 1][2] - xyz_le_sects_rel[i][2],
                        xyz_le_sects_rel[i + 1][1] - xyz_le_sects_rel[i][1],
                    )

                for k, v in {
                    "y_position"         : xyz_le_sects_rel[i][1],
                    "Chord"              : xsec.chord,
                    "xOffset"            : xyz_le_sects_rel[i][0],
                    "Dihedral"           : dihedral,
                    "Twist"              : xsec.twist,
                    "Left_Side_FoilName" : xsec.airfoil.name,
                    "Right_Side_FoilName": xsec.airfoil.name,
                    "x_number_of_panels" : 8,
                    "y_number_of_panels" : 8,
                }.items():
                    subelement = ET.SubElement(sect, k)
                    subelement.text = str(v)

        if elevator is not None:
            wing = elevator
            wingxml = ET.SubElement(plane, "wing")

            xyz_le_root = wing._compute_xyz_of_WingXSec(index=0, x_nondim=0, z_nondim=0)

            for k, v in {
                "Name"      : wing.name,
                "Type"      : "ELEVATOR",
                "Position"  : ",".join([str(x) for x in xyz_le_root]),
                "Tilt_angle": 0.,
                "Symetric"  : wing.symmetric,  # This tag is a typo in XFLR...
                "isFin"     : "false",
                "isSymFin"  : "false",
            }.items():
                subelement = ET.SubElement(wingxml, k)
                subelement.text = str(v)

            sections = ET.SubElement(wingxml, "Sections")

            xyz_le_sects_rel = [
                wing._compute_xyz_of_WingXSec(index=i, x_nondim=0, z_nondim=0) - xyz_le_root
                for i in range(len(wing.xsecs))
            ]

            for i, xsec in enumerate(wing.xsecs):

                sect = ET.SubElement(sections, "Section")

                if i == len(wing.xsecs) - 1:
                    dihedral = 0
                else:
                    dihedral = np.arctan2d(
                        xyz_le_sects_rel[i + 1][2] - xyz_le_sects_rel[i][2],
                        xyz_le_sects_rel[i + 1][1] - xyz_le_sects_rel[i][1],
                    )

                for k, v in {
                    "y_position"         : xyz_le_sects_rel[i][1],
                    "Chord"              : xsec.chord,
                    "xOffset"            : xyz_le_sects_rel[i][0],
                    "Dihedral"           : dihedral,
                    "Twist"              : xsec.twist,
                    "Left_Side_FoilName" : xsec.airfoil.name,
                    "Right_Side_FoilName": xsec.airfoil.name,
                    "x_number_of_panels" : 8,
                    "y_number_of_panels" : 8,
                }.items():
                    subelement = ET.SubElement(sect, k)
                    subelement.text = str(v)

        if fin is not None:
            wing = fin
            wingxml = ET.SubElement(plane, "wing")

            xyz_le_root = wing._compute_xyz_of_WingXSec(index=0, x_nondim=0, z_nondim=0)

            for k, v in {
                "Name"      : wing.name,
                "Type"      : "FIN",
                "Position"  : ",".join([str(x) for x in xyz_le_root]),
                "Tilt_angle": 0.,
                "Symetric"  : "true",  # This tag is a typo in XFLR...
                "isFin"     : "true",
                "isSymFin"  : wing.symmetric,
            }.items():
                subelement = ET.SubElement(wingxml, k)
                subelement.text = str(v)

            sections = ET.SubElement(wingxml, "Sections")

            xyz_le_sects_rel = [
                wing._compute_xyz_of_WingXSec(index=i, x_nondim=0, z_nondim=0) - xyz_le_root
                for i in range(len(wing.xsecs))
            ]

            for i, xsec in enumerate(wing.xsecs):

                sect = ET.SubElement(sections, "Section")

                if i == len(wing.xsecs) - 1:
                    dihedral = 0
                else:
                    dihedral = np.arctan2d(
                        xyz_le_sects_rel[i + 1][1] - xyz_le_sects_rel[i][1],
                        xyz_le_sects_rel[i + 1][2] - xyz_le_sects_rel[i][2],
                    )

                for k, v in {
                    "y_position"         : xyz_le_sects_rel[i][2],
                    "Chord"              : xsec.chord,
                    "xOffset"            : xyz_le_sects_rel[i][0],
                    "Dihedral"           : dihedral,
                    "Twist"              : xsec.twist,
                    "Left_Side_FoilName" : xsec.airfoil.name,
                    "Right_Side_FoilName": xsec.airfoil.name,
                    "x_number_of_panels" : 8,
                    "y_number_of_panels" : 8,
                }.items():
                    subelement = ET.SubElement(sect, k)
                    subelement.text = str(v)

        ### Indents the XML file properly
        def indent(elem, level=0):
            i = "\n" + level * "  "
            if len(elem):
                if not elem.text or not elem.text.strip():
                    elem.text = i + "  "
                if not elem.tail or not elem.tail.strip():
                    elem.tail = i
                for elem in elem:
                    indent(elem, level + 1)
                if not elem.tail or not elem.tail.strip():
                    elem.tail = i
            else:
                if level and (not elem.tail or not elem.tail.strip()):
                    elem.tail = i

        indent(root)

        xml_string = ET.tostring(
            root,
            encoding="UTF-8",
            xml_declaration=True
        ).decode()

        with open(filename, "w+") as f:
            f.write(xml_string)

        return xml_string

    def export_OpenVSP_vspscript(self,
                                 filename: Union[Path, str],
                                 ) -> str:
        """
        Exports the airplane geometry to a `*.vspscript` file compatible with OpenVSP. To import the `.vspscript`
        file into OpenVSP:

        Open OpenVSP, then File -> Run Script -> Select the `.vspscript` file.

        Args:
            filename: The filename to export to, given as a string or Path. Should include the ".vspscript" extension.

        Returns: A string of the file contents, and also saves the file to the specified filename
        """
        from aerosandbox.geometry.openvsp_io.asb_to_openvsp.airplane_vspscript_generator import generate_airplane

        vspscript_code = generate_airplane(self)

        with open(filename, "w+") as f:
            f.write(vspscript_code)

        return vspscript_code


if __name__ == '__main__':
    import aerosandbox as asb
    # import aerosandbox.numpy as np
    import aerosandbox.tools.units as u


    def ft(feet, inches=0):  # Converts feet (and inches) to meters
        return feet * u.foot + inches * u.inch


    naca2412 = asb.Airfoil("naca2412")
    naca0012 = asb.Airfoil("naca0012")

    airplane = Airplane(
        name="Cessna 152",
        wings=[
            asb.Wing(
                name="Wing",
                xsecs=[
                    asb.WingXSec(
                        xyz_le=[0, 0, 0],
                        chord=ft(5, 4),
                        airfoil=naca2412
                    ),
                    asb.WingXSec(
                        xyz_le=[0, ft(7), ft(7) * np.sind(1)],
                        chord=ft(5, 4),
                        airfoil=naca2412,
                        control_surfaces=[
                            asb.ControlSurface(
                                name="aileron",
                                symmetric=False,
                                hinge_point=0.8,
                                deflection=0
                            )
                        ]
                    ),
                    asb.WingXSec(
                        xyz_le=[
                            ft(4, 3 / 4) - ft(3, 8 + 1 / 2),
                            ft(33, 4) / 2,
                            ft(33, 4) / 2 * np.sind(1)
                        ],
                        chord=ft(3, 8 + 1 / 2),
                        airfoil=naca0012
                    )
                ],
                symmetric=True
            ),
            asb.Wing(
                name="Horizontal Stabilizer",
                xsecs=[
                    asb.WingXSec(
                        xyz_le=[0, 0, 0],
                        chord=ft(3, 8),
                        airfoil=naca0012,
                        twist=-2,
                        control_surfaces=[
                            asb.ControlSurface(
                                name="elevator",
                                symmetric=True,
                                hinge_point=0.75,
                                deflection=0
                            )
                        ]
                    ),
                    asb.WingXSec(
                        xyz_le=[ft(1), ft(10) / 2, 0],
                        chord=ft(2, 4 + 3 / 8),
                        airfoil=naca0012,
                        twist=-2
                    )
                ],
                symmetric=True
            ).translate([ft(13, 3), 0, ft(-2)]),
            asb.Wing(
                name="Vertical Stabilizer",
                xsecs=[
                    asb.WingXSec(
                        xyz_le=[ft(-5), 0, 0],
                        chord=ft(8, 8),
                        airfoil=naca0012,
                    ),
                    asb.WingXSec(
                        xyz_le=[ft(0), 0, ft(1)],
                        chord=ft(3, 8),
                        airfoil=naca0012,
                        control_surfaces=[
                            asb.ControlSurface(
                                name="rudder",
                                hinge_point=0.75,
                                deflection=0
                            )
                        ]
                    ),
                    asb.WingXSec(
                        xyz_le=[ft(0, 8), 0, ft(5)],
                        chord=ft(2, 8),
                        airfoil=naca0012,
                    ),
                ]
            ).translate([ft(16, 11) - ft(3, 8), 0, ft(-2)])
        ],
        fuselages=[
            asb.Fuselage(
                xsecs=[
                    asb.FuselageXSec(
                        xyz_c=[0, 0, ft(-1)],
                        radius=0,
                    ),
                    asb.FuselageXSec(
                        xyz_c=[0, 0, ft(-1)],
                        radius=ft(1.5),
                        shape=3,
                    ),
                    asb.FuselageXSec(
                        xyz_c=[ft(3), 0, ft(-0.85)],
                        radius=ft(1.7),
                        shape=7,
                    ),
                    asb.FuselageXSec(
                        xyz_c=[ft(5), 0, ft(0)],
                        radius=ft(2.7),
                        shape=7,
                    ),
                    asb.FuselageXSec(
                        xyz_c=[ft(10, 4), 0, ft(0.3)],
                        radius=ft(2.3),
                        shape=7,
                    ),
                    asb.FuselageXSec(
                        xyz_c=[ft(21, 11), 0, ft(0.8)],
                        radius=ft(0.3),
                        shape=3,
                    ),
                ]
            ).translate([ft(-5), 0, ft(-3)])
        ]
    )

    airplane.draw_three_view()
    # airplane.export_XFLR5_xml("test.xml", mass_props=asb.MassProperties(mass=1, Ixx=1, Iyy=1, Izz=1))



"""A setuptools based setup module.

See:
https://packaging.python.org/guides/distributing-packages-using-setuptools/
https://github.com/pypa/sampleproject
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
from pathlib import Path

# Get the long description from the README file
README_path = Path(__file__).parent / "README.md"
with open(README_path, encoding='utf-8') as f:
    long_description = f.read()

### Get the version number dynamically
init_path = Path(__file__).parent / "aerosandbox" / "__init__.py"

with open(init_path) as f:
    for line in f:
        if line.startswith("__version__"):
            version = line.split("=")[1].strip().strip('"').strip("'")
            break
    else:
        raise RuntimeError("Unable to find version string.")

### Do the setup
setup(
    name='AeroSandbox',
    author='Peter Sharpe',
    version=version,
    description='AeroSandbox is a Python package that helps you design and optimize aircraft and other engineered systems.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://peterdsharpe.github.io/AeroSandbox/',
    author_email='pds@mit.edu',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Physics',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
    ],
    keywords='aerodynamics airplane cfd mdo mdao aircraft design aerospace optimization automatic differentiation structures propulsion',
    packages=find_packages(exclude=['docs', 'media', 'studies', 'tutorial']),
    python_requires='>=3.8',
    install_requires=[
        'numpy >= 1.20.0, <2.0a0',
        'scipy >= 1.7.0',
        'casadi >= 3.6.4',
        'pandas >= 2',
        'matplotlib >= 3.7.0',
        'seaborn >= 0.11',
        'tqdm >= 4',
        'sortedcontainers >= 2',
        'dill >= 0.3',
        'neuralfoil >= 0.2.1, <0.3.0'
    ],
    extras_require={
        "full": [
            'plotly >= 5',
            'pyvista >= 0.31',
            'trimesh >= 3',
            'sympy >= 1',
            'shapely >= 2',
        ],
        "test": [
            'pytest<=8.0.2',
            'nbval',
            'shapely >= 2',  # Required for some airfoil geometry tests
            'openpyxl', # Required for some data imports in tutorials
        ],
        "docs": [
            'sphinx',
            'furo',
            'sphinx-autoapi',
        ],
    },
    include_package_data=True,
    package_data={
        'Airfoil database': ['*.dat'],  # include all airfoil *.dat files
    },
    project_urls={  # Optional
        'Source'     : 'https://github.com/peterdsharpe/AeroSandbox',
        'Bug Reports': 'https://github.com/peterdsharpe/AeroSandbox/issues',
    },
)



import aerosandbox.numpy as np
from aerosandbox.geometry.airplane import Airplane
from textwrap import indent, dedent
from aerosandbox.geometry.openvsp_io.asb_to_openvsp import _utilities
from aerosandbox.geometry.openvsp_io.asb_to_openvsp.wing_vspscript_generator import generate_wing
from aerosandbox.geometry.openvsp_io.asb_to_openvsp.fuselage_vspscript_generator import generate_fuselage
from aerosandbox.geometry.openvsp_io.asb_to_openvsp.propulsor_vspscript_generator import generate_propulsor


def generate_airplane(
        airplane: Airplane,
        include_main=True,
) -> str:
    """
    Generates a VSPScript file for an Airplane object.

    Args:
        airplane: The Airplane object to generate a VSPScript file for.

    Returns: A VSPScript file as a string.

    """
    script = ""

    for wing in airplane.wings:
        script += f"""\
{{
{indent(generate_wing(wing, include_main=False), "    ")}
}}
"""

    for fuselage in airplane.fuselages:
        script += f"""\
{{
{indent(generate_fuselage(fuselage, include_main=False), "    ")}
}}
"""

    for propulsor in airplane.propulsors:
        script += f"""\
{{
{indent(generate_propulsor(propulsor, include_main=False), "    ")}
}}
"""

    if include_main:
        script = _utilities.wrap_script(script)

    return script


if __name__ == '__main__':

    from aerosandbox.geometry.wing import Wing, WingXSec
    from aerosandbox.geometry.airfoil.airfoil import Airfoil

    wing = Wing(
        name="Main Wing",
        symmetric=True,
        xsecs=[
            WingXSec(
                xyz_le=[0.5, 0, 0],
                chord=1.1,
                twist=5,
                airfoil=Airfoil(name="dae11")
            ),
            WingXSec(
                xyz_le=[1, 2, 0],
                chord=0.9,
                twist=5,
                airfoil=Airfoil(name="NACA4412")
            ),
            WingXSec(
                xyz_le=[2, 5, 0],
                chord=0.5,
                twist=0,
                airfoil=Airfoil(name="NACA3412")
            ),
            WingXSec(
                xyz_le=[2.5, 5.5, 1],
                chord=0.25,
                twist=0,
                airfoil=Airfoil(name="NACA2412")
            )
        ]
    )

    from aerosandbox.geometry.fuselage import Fuselage, FuselageXSec

    af = Airfoil("ls013")
    x = np.sinspace(0, 1, 10)

    fuse = Fuselage(
        name="Fuse",
        xsecs=[
            FuselageXSec(
                xyz_c=[xi, 0, 0.05 * xi ** 2],
                width=2 * af.local_thickness(xi),
                height=af.local_thickness(xi),
                shape=4
            )
            for xi in x
        ]
    )

    from aerosandbox.geometry.propulsor import Propulsor

    prop = Propulsor(
        name="Prop",
        xyz_c=np.array([0.5, 0.4, 0.3]),
        xyz_normal=np.array([1, 0.2, 0.3]),
        radius=0.6,
    )

    airplane = Airplane(
        name="Aircraft",
        wings=[wing],
        fuselages=[fuse],
        propulsors=[prop]
    )

    print(generate_airplane(airplane))

    with open("test.vspscript", "w") as f:
        f.write(generate_airplane(airplane))



import aerosandbox.numpy as np
import aerosandbox.tools.units as u
from typing import Dict


def modified_DAPCA_IV_production_cost_analysis(
        design_empty_weight: float,
        design_maximum_airspeed: float,
        n_airplanes_produced: int,
        n_engines_per_aircraft: int,
        cost_per_engine: float,
        cost_avionics_per_airplane: float,
        n_pax: int,
        cpi_relative_to_2012_dollars: float = 1.327,  # updated for 2024
        n_flight_test_aircraft: int = 4,
        is_cargo_airplane: bool = False,
        primary_structure_material: str = "aluminum",
        per_passenger_cost_model: str = "general_aviation",
        engineering_wrap_rate_2012_dollars: float = 115.,
        tooling_wrap_rate_2012_dollars: float = 118.,
        quality_control_wrap_rate_2012_dollars: float = 108.,
        manufacturing_wrap_rate_2012_dollars: float = 98.,
) -> Dict[str, float]:
    """
    Computes the cost of an aircraft in present-day dollars, using the Modified DAPCA IV cost model.

    Be sure to adjust `cpi_relative_to_2012_dollars` to the current values in order to accurately model inflation.

    The DAPCA IV cost model is a statistical regression of historical aircraft cost data. It provides reasonable
    results for most classes of aircraft, including transports, fighters, bombers, and even GA and UAV aircraft with
    suitable adjustments.

    It was created by the RAND Corporation.

    The Modified DAPCA IV cost model is a modification of the DAPCA IV cost model that includes additional cost
    estimates for engine cost (as the original DAPCA model assumes that this is known).

    See Raymer, Aircraft Design: A Conceptual Approach, 5th Edition, Section 18.4.2 pg. 711 for more information.

    Args:
        design_empty_weight: The design empty weight of the entire aircraft, in kg.

        design_maximum_airspeed: The design maximum airspeed of the aircraft, in m/s.

        n_airplanes_produced: The number of airplanes to be produced.

        n_engines_per_aircraft: The number of engines per aircraft.

        cost_per_engine: The cost of each engine, in present-day dollars.

        cost_avionics_per_airplane: The cost of avionics per airplane, in present-day dollars.

        n_pax: The number of passengers.

        cpi_relative_to_2012_dollars: The consumer price index at the present day divided by the consumer price index
        in 2012, seasonally-adjusted.

            To quickly find this, use data from the St. Louis Federal Reserve. Below is the CPI, normalized to 2012.
            https://fred.stlouisfed.org/graph/?g=10PU0

            For example, in 2022, one would use 1.275.

        n_flight_test_aircraft: The number of flight test aircraft. Typically 2 to 6.

        is_cargo_airplane: Whether the airplane is a cargo airplane. If so, the quality control cost is lower.

        primary_structure_material: The primary structure material. Options are:
            - "aluminum"
            - "carbon_fiber"
            - "fiberglass"
            - "steel"
            - "titanium"

        per_passenger_cost_model: The per-passenger cost model. Options are:
            - "general_aviation": General aviation aircraft, such as Cessna 172s.
            - "jet_transport": Jet transport aircraft, such as Boeing 737s.
            - "regional_transport": Regional transport aircraft, such as Embraer E175s.

        engineering_wrap_rate_2012_dollars: The engineering wrap rate in 2012 dollars.

        tooling_wrap_rate_2012_dollars: The tooling wrap rate in 2012 dollars.

        quality_control_wrap_rate_2012_dollars: The quality control wrap rate in 2012 dollars.

        manufacturing_wrap_rate_2012_dollars: The manufacturing wrap rate in 2012 dollars.

    Returns:

        A dictionary of costs required to produce all `n_airplanes_produced` airplanes, in present-day dollars.

        Keys and values are as follows:

            - "engineering_labor": Engineering labor cost.

            - "tooling_labor": Tooling labor cost.

            - "manufacturing_labor": Manufacturing labor cost.

            - "quality_control_labor": Quality control labor cost.

            - "development_support": Development support cost. From Raymer: "Includes fabrication of mockups, iron-bird subsystem
            simulators, structural test articles, and other test articles."

            - "flight_test": Flight test cost. From Raymer: "Includes all costs incurred to demonstrate airworthiness
            for civil certification or Mil-Spec compliance except for the costs of the flight-test aircraft
            themselves. Costs for the flight-test aircraft are included in the total production-run cost estimation.
            Includes planning, instrumentation, flight operations, data reduction, and engineering and manufacturing
            support of flight testing."

            - "manufacturing_materials": Manufacturing materials cost. From Raymer: "Includes all raw materials and
            purchased hardware and equipment."

            - "engines": Engine cost.

            - "avionics": Avionics cost.

            - "total": Total cost. (Sum of all other costs above.)

    """
    # Abbreviated constants for readability
    W = design_empty_weight  # kg
    V = design_maximum_airspeed / u.kph  # km/hour
    Q = n_airplanes_produced

    ### Estimate labor hours
    hours = dict()

    hours["engineering"] = 5.18 * W ** 0.777 * V ** 0.894 * Q ** 0.163
    hours["tooling"] = 7.22 * W ** 0.777 * V ** 0.696 * Q ** 0.263
    hours["manufacturing"] = 10.5 * W ** 0.82 * V ** 0.484 * Q ** 0.641
    hours["quality_control"] = hours["manufacturing"] * (0.076 if is_cargo_airplane else 0.133)

    ### Account for materials difficulties
    if primary_structure_material == "aluminum":
        materials_hourly_multiplier = 1.0
    elif primary_structure_material == "carbon_fiber":
        materials_hourly_multiplier = (1.1 + 1.8) / 2
    elif primary_structure_material == "fiberglass":
        materials_hourly_multiplier = (1.1 + 1.2) / 2
    elif primary_structure_material == "steel":
        materials_hourly_multiplier = (1.5 + 2.0) / 2
    elif primary_structure_material == "titanium":
        materials_hourly_multiplier = (1.1 + 1.8) / 2
    else:
        raise ValueError("Invalid value of `primary_structure_material`.")

    hours = {
        k: v * materials_hourly_multiplier
        for k, v in hours.items()
    }

    ### Convert labor hours to labor costs in 2012 dollars
    costs_2012_dollars = dict()

    costs_2012_dollars["engineering_labor"] = hours["engineering"] * engineering_wrap_rate_2012_dollars
    costs_2012_dollars["tooling_labor"] = hours["tooling"] * tooling_wrap_rate_2012_dollars
    costs_2012_dollars["manufacturing_labor"] = hours["manufacturing"] * manufacturing_wrap_rate_2012_dollars
    costs_2012_dollars["quality_control_labor"] = hours["quality_control"] * quality_control_wrap_rate_2012_dollars

    costs_2012_dollars["development_support"] = 67.4 * W ** 0.630 * V ** 1.3
    costs_2012_dollars["flight_test"] = 1947 * W ** 0.325 * V ** 0.822 * n_flight_test_aircraft ** 1.21
    costs_2012_dollars["manufacturing_materials"] = 31.2 * W ** 0.921 * V ** 0.621 * Q ** 0.799

    ### Add in the per-passenger cost for aircraft interiors:
    # Seats, luggage bins, closets, lavatories, insulation, ceilings, floors, walls, etc.
    # Costs are from Raymer, Aircraft Design: A Conceptual Approach, 5th edition. Section 18.4.2, page 715.
    if per_passenger_cost_model == "general_aviation":
        costs_2012_dollars["aircraft_interiors"] = n_airplanes_produced * n_pax * 850
    elif per_passenger_cost_model == "jet_transport":
        costs_2012_dollars["aircraft_interiors"] = n_airplanes_produced * n_pax * 3500
    elif per_passenger_cost_model == "regional_transport":
        costs_2012_dollars["aircraft_interiors"] = n_airplanes_produced * n_pax * 1700
    else:
        raise ValueError(f"Invalid value of `per_passenger_cost_model`!")

    ### Convert all costs to present-day dollars
    costs = {
        k: v * cpi_relative_to_2012_dollars
        for k, v in costs_2012_dollars.items()
    }

    ### Add the engine(s) and avionics costs
    costs["engines"] = cost_per_engine * n_engines_per_aircraft * n_airplanes_produced
    costs["avionics"] = cost_avionics_per_airplane * n_airplanes_produced

    ### Total all costs and return
    costs["total"] = sum(costs.values())

    return costs


def electric_aircraft_direct_operating_cost_analysis(
        production_cost_per_airframe: float,
        nominal_cruise_airspeed: float,
        nominal_mission_range: float,
        battery_capacity: float,
        num_passengers_nominal: int,
        num_crew: int = 1,
        battery_fraction_used_on_nominal_mission: float = 0.8,
        typical_passenger_utilization: float = 0.8,
        flight_hours_per_year: float = 1200,
        airframe_lifetime_years: float = 20,
        airframe_eol_resale_value_fraction: float = 0.4,
        battery_cost_per_kWh_capacity: float = 500.0,
        battery_cycle_life: float = 1500,
        real_interest_rate: float = 0.04,
        electricity_cost_per_kWh: float = 0.145,
        annual_expenses_per_crew: float = 100000 * 1.5,
        ascent_time: float = 0.2 * u.hour,
        descent_time: float = 0.2 * u.hour,
) -> Dict[str, float]:
    """
    Estimates the overall operating cost of an electric aircraft. Includes both direct and indirect operating costs.

    Here, direct operating costs (DOC) are taken to include the following costs:

    - Airframe depreciation
    - Airframe financing
    - Insurance
    - Maintenance
    - Battery replacement
    - Energy costs (here, electricity)
    - Cockpit and cabin crew costs
    - Airport landing, terminal, and handling fees

    Any costs that are not included here are considered indirect costs. These indirect costs would include,
    but are not limited to: advertisement, administrative costs, depreciation of non-airframe assets, and taxes.

    Airframe maintenance costs are estimated from:
        Moore, et al., "Unlocking Low-Cost Regional Air Mobility through...", AIAA Aviation 2023.

    Airport fees estimated for the Phoenix-Mesa Gateway Airport, due to public availability of the fee schedule:
    https://www.gatewayairport.com/documents/documentlibrary/wgaa%20organizational%20documents/airport%20rates%20charges%20-%20effective%20march%201,%202017.pdf

    Args:

        production_cost_per_airframe: The cost to produce a single airframe, in present-day dollars. May be estimated
            using the `modified_DAPCA_IV_production_cost_analysis()` function.

        nominal_cruise_airspeed: The nominal cruise airspeed of the aircraft, in m/s.

        nominal_mission_range: The nominal mission range of the aircraft, in meters.

        battery_capacity: The total capacity of the battery, in Joules.

        num_passengers_nominal: The number of passengers that the aircraft is designed to carry.

        num_crew: The number of crew members required to operate the aircraft.

        battery_fraction_used_on_nominal_mission: The fraction of the battery's capacity that is used on the nominal
            mission.

        typical_passenger_utilization: The fraction of the aircraft's passenger capacity that is typically utilized.

        flight_hours_per_year: The number of flight hours per year that the aircraft is expected to fly.

        airframe_lifetime_years: The number of years that the airframe is expected to last. After this time, the airframe
            is assumed to be sold at some lower reasle value.

        airframe_eol_resale_value_fraction: The expect resale value of the airframe at the end of its lifetime,
            expressed as a fraction of the airframe's production cost.

        battery_cost_per_kWh_capacity: The replacement cost of the battery pack, per kWh of capacity, in present-day
            dollars. Note that this is a pack-level cost (as opposed to cell-level), so includes the cost of the
            battery management system, cooling system, fire-suppressing foam, etc. Should include the labor cost to
            replace the battery pack as well.

        battery_cycle_life: The number of charge/discharge cycles that the battery is expected to last before full
            replacement is required.

        real_interest_rate: The real interest rate per year. This is the interest rate minus the inflation rate. This is
            used to calculate the present-day value of future costs (e.g., airframe financing).

        electricity_cost_per_kWh: The cost of electricity, per kWh, in present-day dollars.

        annual_expenses_per_crew: The annual expenses per crew member, in present-day dollars. Should include the total
            burdened cost of the crew member, including salary, benefits, and other expenses.

        ascent_time: The time required to ascend to cruise altitude, in seconds.

        descent_time: The time required to descend from cruise altitude, in seconds.

    Returns:
        A dictionary of operating costs per passenger-mile, in present-day dollars, with the following keys:

            * "airframe_depreciation"
            * "airframe_financing"
            * "insurance"
            * "airframe_maintenance"
            * "propulsion_maintenance"
            * "battery_replacement"
            * "energy"
            * "crew"
            * "airport_landing_fees"
            * "airport_terminal_fees"
            * "airport_parking_fees"
            * "airport_passenger_facility_charge"
            * "indirect"

        One key, "total", is also included, which is the sum of all of the above costs. Once again, this is
            expressed in units of present-day dollars per passenger-mile.

    """

    ### Calculate per-mission parameters
    ascent_and_descent_airspeed = 0.75 * nominal_cruise_airspeed
    ascent_and_descent_time = np.softmin(
        ascent_time + descent_time,
        nominal_mission_range / ascent_and_descent_airspeed,
        softness=2 * u.minute,
    )
    ascent_and_descent_distance = ascent_and_descent_airspeed * ascent_and_descent_time
    cruise_distance = nominal_mission_range - ascent_and_descent_distance
    cruise_time = cruise_distance / nominal_cruise_airspeed
    mission_time = ascent_and_descent_time + cruise_time

    ### Calculate annual utilization parameters
    flights_per_year = flight_hours_per_year / (mission_time / u.hour)

    ### Begin calculating costs on a per-flight basis
    costs_per_flight = dict()  # Lists the cost per flight for each cost category

    ### Airframe depreciation costs
    num_airframe_flights_lifetime = flights_per_year * airframe_lifetime_years
    net_value_per_airframe_over_lifetime = (
            production_cost_per_airframe  # Production cost
            - (production_cost_per_airframe * airframe_eol_resale_value_fraction)  # End-of-life resale value
    )
    costs_per_flight["airframe_depreciation"] = (
            net_value_per_airframe_over_lifetime / num_airframe_flights_lifetime
    )

    ### Airframe financing cost
    airframe_financing_lifetime_cost = production_cost_per_airframe * (
            np.exp(real_interest_rate * airframe_lifetime_years) - 1
    )
    costs_per_flight["airframe_financing"] = (
            airframe_financing_lifetime_cost / num_airframe_flights_lifetime
    )

    ### Insurance cost
    insurance_cost_per_year = production_cost_per_airframe * (
            0.025  # Base rate of 2% on average, adjusted higher for perceived higher risk of new electric technology
            * ((num_passengers_nominal + num_crew + 1) / 11) ** 0.5  # Adj. for number of souls on board; but sublinear
            * (flight_hours_per_year / 1200) ** 0.5  # Normalizing to industry average; but again, sublinear
    )
    costs_per_flight["insurance"] = insurance_cost_per_year / flights_per_year

    ### Airframe maintenance cost
    # maintenance_cost_per_year = production_cost_per_airframe * 0.04
    # costs_per_flight["maintenance"] = maintenance_cost_per_year / flights_per_year
    costs_per_flight["airframe_maintenance"] = (production_cost_per_airframe / 3e6) * (
            65 * (mission_time / u.hour) +
            (65)  # per cycle
    )

    ### Propulsion maintenance cost
    costs_per_flight["propulsion_maintenance"] = (production_cost_per_airframe / 3e6) * (
            (58 * (mission_time / u.hour)) +
            (50)  # per cycle
    )

    ### Battery replacement cost
    battery_capacity_kWh = battery_capacity / (u.kilo * u.watt_hour)
    battery_cost = battery_capacity_kWh * battery_cost_per_kWh_capacity
    costs_per_flight["battery_replacement"] = battery_cost / battery_cycle_life

    ### Energy Cost
    electric_energy_per_flight = battery_capacity * battery_fraction_used_on_nominal_mission
    electric_energy_per_flight_kWh = electric_energy_per_flight / (u.kilo * u.watt_hour)
    costs_per_flight["energy"] = electric_energy_per_flight_kWh * electricity_cost_per_kWh

    ### Crew cost
    costs_per_flight["crew"] = (
            num_crew *
            annual_expenses_per_crew /
            flights_per_year
    )

    ### Airport landing fees
    estimated_max_gross_landing_weight = (  # Model is very approximate, because landing fees are small
            production_cost_per_airframe / 266  # typical cost per kg empty weight
            * 1.2  # Rough ratio of landing weight to empty weight
    )
    costs_per_flight["airport_landing_fees"] = (
            1.20 * estimated_max_gross_landing_weight / 1000
    )

    ### Airport terminal use fees
    costs_per_flight["airport_terminal_fees"] = (
        50  # Cost per turn-around
    )

    ### Airport aircraft parking fees
    costs_per_flight["airport_parking_fees"] = (
        35  # Cost per turn-around
    )

    ### Airport passenger facility charge
    costs_per_flight["airport_passenger_facility_charge"] = (
            4.50 * num_passengers_nominal
    )

    ### Indirect Costs
    costs_per_flight["indirect"] = 0.2 * sum(costs_per_flight.values())

    # Cost per passenger mile
    n_passengers_actual = num_passengers_nominal * typical_passenger_utilization
    passenger_miles_per_flight = n_passengers_actual * (nominal_mission_range / u.mile)
    costs_per_flight["total"] = sum(costs_per_flight.values())

    # Return same dictionary as before
    costs_per_paxmi = {
        k: v / passenger_miles_per_flight
        for k, v in costs_per_flight.items()
    }

    return costs_per_paxmi

if __name__ == '__main__':
    res = electric_aircraft_direct_operating_cost_analysis(
        production_cost_per_airframe=3.0e6,
        nominal_cruise_airspeed=250 * u.knot,
        nominal_mission_range=150 * u.mile,
        battery_capacity=800e3 * u.watt_hour,
        num_passengers_nominal=9,
    )
    import pandas as pd

    print(pd.Series(res))



from aerosandbox.common import ExplicitAnalysis
import aerosandbox.numpy as np
import subprocess
from pathlib import Path
from aerosandbox.geometry import Airfoil
from typing import Union, List, Dict
import tempfile
import warnings
import os
import re


class XFoil(ExplicitAnalysis):
    """

    An interface to XFoil, a 2D airfoil analysis tool developed by Mark Drela at MIT.

    Requires XFoil to be on your computer; XFoil is available here: https://web.mit.edu/drela/Public/web/xfoil/

    It is recommended (but not required) that you add XFoil to your system PATH environment variable such that it can
    be called with the command `xfoil`. If this is not the case, you need to specify the path to your XFoil
    executable using the `xfoil_command` argument of the constructor.

    Usage example:

    >>> xf = XFoil(
    >>>     airfoil=Airfoil("naca2412").repanel(n_points_per_side=100),
    >>>     Re=1e6,
    >>> )
    >>>
    >>> result_at_single_alpha = xf.alpha(5)
    >>> result_at_several_CLs = xf.cl([0.5, 0.7, 0.8, 0.9])
    >>> result_at_multiple_alphas = xf.alpha([3, 5, 60]) # Note: if a result does not converge (such as the 60 degree case here), it will not be included in the results.

    """

    # Defines an exception to throw if XFoil fails externally
    class XFoilError(Exception):
        pass

    def __init__(self,
                 airfoil: Airfoil,
                 Re: float = 0.,
                 mach: float = 0.,
                 n_crit: float = 9.,
                 xtr_upper: float = 1.,
                 xtr_lower: float = 1.,
                 hinge_point_x: float = 0.75,
                 full_potential: bool = False,
                 max_iter: int = 100,
                 xfoil_command: str = "xfoil",
                 xfoil_repanel: bool = True,
                 xfoil_repanel_n_points: int = 279,
                 include_bl_data: bool = False,
                 verbose: bool = False,
                 timeout: Union[float, int, None] = 30,
                 working_directory: Union[Path, str] = None,
                 ):
        """
        Interface to XFoil. Compatible with both XFoil v6.xx (public) and XFoil v7.xx (private, contact Mark Drela at
        MIT for a copy.)

        Args:

            airfoil: The airfoil to analyze. Should be an AeroSandbox Airfoil object.

            Re: The chord-referenced Reynolds number. Set this to 0 to run in inviscid mode.

            mach: The freestream Mach number. Note that XFoil 6.xx uses the Karman-Tsien compressibility correction,
                which breaks down once supersonic flow is present (i.e., past M_crit). XFoil 7.xx has a full-potential
                solver that is theoretically-valid for weak shocks (perhaps up to M_crit + 0.05 or so).

            n_crit: The critical Tollmein-Schlichting wave amplification factor, as part of the "e^n" transition
                criterion. This is a measure of freestream turbulence and surface roughness. The following reference conditions
                are given in the XFoil documentation:

                - sailplane:                12-14
                - motorglider:              11-13
                - clean wind tunnel:        10-12
                - average wind tunnel:      9 (default)
                - dirty wind tunnel:        4-8

            xtr_upper: The upper-surface forced transition location [x/c], where the boundary layer will be
                automatically tripped to turbulent. Set to 1 to disable forced transition (default). Note that if the
                Reynolds number is sufficiently low, it is possible for the flow to re-laminarize after being tripped.

            xtr_lower: The lower-surface forced transition location [x/c], where the boundary layer will be
                automatically tripped to turbulent. Set to 1 to disable forced transition (default). Note that if the
                Reynolds number is sufficiently low, it is possible for the flow to re-laminarize after being tripped.

            hinge_point_x: The x/c location of the hinge point. This is used to calculate the hinge moment. If this is
                None, the hinge moment is not calculated.

            full_potential: If this is set True, it will turn full-potential mode on. Note that full-potential mode
                is only available in XFoil v7.xx or higher. (Unless you have specifically gone through the trouble of
                acquiring a copy of XFoil v7.xx you likely have v6.xx. Version 7.xx is not publicly distributed as of
                2023; contact Mark Drela at MIT for a copy.) Note that if you enable this flag with XFoil v6.xx,
                you'll likely get an error (no output file generated).

            max_iter: How many iterations should we let XFoil do?

            xfoil_command: The command-line argument to call XFoil, given as a string or a Path-like object.

                * If XFoil is on your system PATH, then you can just leave this as "xfoil".

                * If XFoil is not on your system PATH, then you should provide a filepath to the XFoil executable.

                Note that XFoil is not on your PATH by default. To tell if XFoil is not on your system PATH,
                open up a terminal and type "xfoil".

                    * If the XFoil menu appears, it's on your PATH.

                    * If you get something like "'xfoil' is not recognized as an internal or external command..." or
                    "Command 'xfoil' not found, did you mean...", then it is not on your PATH and you'll need to
                    specify the location of your XFoil executable as a string.

                To add XFoil to your path, modify your system's environment variables. (Google how to do this for
                your OS.)

            xfoil_repanel: Controls whether to allow XFoil to repanel your airfoil using its internal methods (PANE,
                with default settings, 160 nodes). Boolean, defaults to True.

            xfoil_repanel_n_points: If `xfoil_repanel` is True, this controls the number of points to repanel the
                airfoil to within XFoil. Defaults to 279, which is the highest number of panel points before XFoil's
                `IWX` array overfills and starts trimming wake points.

            include_bl_data: Controls whether or not to include boundary layer data in the output. If this is True,
                the functions `alpha()` and `cl()` will return a dictionary with an additional key, "bl_data",
                which contains the boundary layer data in the form of a pandas DataFrame. Results in slightly higher
                runtime, mostly due to file I/O bottleneck. Defaults to False.

            verbose: Controls whether or not XFoil output is printed to command line. Defaults to False.

            timeout: Controls how long any individual XFoil run (i.e. alpha sweep) is allowed to run before the
                process is killed. Given in units of seconds. To disable timeout, set this to None.

            working_directory: Controls which working directory is used for the XFoil input and output files. By
                default, this is set to a TemporaryDirectory that is deleted after the run. However, you can set it to
                somewhere local for debugging purposes.

        """
        if mach >= 1:
            raise ValueError("XFoil will terminate if a supersonic freestream Mach number is given.")

        self.airfoil = airfoil
        self.Re = Re
        self.mach = mach
        self.n_crit = n_crit
        self.xtr_upper = xtr_upper
        self.xtr_lower = xtr_lower
        self.hinge_point_x = hinge_point_x
        self.full_potential = full_potential
        self.max_iter = max_iter
        self.xfoil_command = xfoil_command
        self.xfoil_repanel = xfoil_repanel
        self.xfoil_repanel_n_points = xfoil_repanel_n_points
        self.include_bl_data = include_bl_data
        self.verbose = verbose
        self.timeout = timeout

        if working_directory is None:
            self.working_directory = None
        else:
            self.working_directory = Path(working_directory)

    def __repr__(self):
        return f"XFoil(airfoil={self.airfoil}, Re={self.Re}, mach={self.mach}, n_crit={self.n_crit})"

    def _default_keystrokes(self,
                            airfoil_filename: str,
                            output_filename: str,
                            ) -> List[str]:
        """
        Returns a list of XFoil keystrokes that are common to all XFoil runs.

        Returns:
            A list of strings, each of which is a single XFoil keystroke to be followed by <enter>.
        """
        run_file_contents = []

        # Disable graphics
        run_file_contents += [
            "plop",
            "g",
            "w 0.05",
            "",
        ]

        # Load the airfoil
        run_file_contents += [
            f"load {airfoil_filename}",
        ]

        if self.xfoil_repanel:
            run_file_contents += [
                "ppar",
                f"n {self.xfoil_repanel_n_points}",
                "",
                "",
                "",
                # "pane",
            ]

        # Enter oper mode
        run_file_contents += [
            "oper",
        ]

        # Handle Re
        if self.Re != 0:
            run_file_contents += [
                f"v {self.Re:.8g}",
            ]

        # Handle mach
        run_file_contents += [
            f"m {self.mach:.8g}",
        ]

        # Handle hinge moment
        run_file_contents += [
            "hinc",
            f"fnew {float(self.hinge_point_x):.8g} {float(self.airfoil.local_camber(self.hinge_point_x)):.8g}",
            "fmom",
        ]

        if self.full_potential:
            run_file_contents += [
                "full",
                "fpar",
                f"i {self.max_iter}",
                "",
            ]

        # Handle iterations
        run_file_contents += [
            f"iter {self.max_iter}",
        ]

        # Handle trips and ncrit
        if not (self.xtr_upper == 1 and self.xtr_lower == 1 and self.n_crit == 9):
            run_file_contents += [
                "vpar",
                f"xtr {self.xtr_upper:.8g} {self.xtr_lower:.8g}",
                f"n {self.n_crit:.8g}",
                "",
            ]

        # Set polar accumulation
        run_file_contents += [
            "pacc",
            f"{output_filename}",
            "",
        ]

        # Include more data in polar
        run_file_contents += [
            "cinc"  # include minimum Cp
        ]

        return run_file_contents

    def _run_xfoil(self,
                   run_command: str,
                   read_bl_data_from: str = None,
                   ) -> Dict[str, np.ndarray]:
        """
        Private function to run XFoil.

        Args: run_command: A string with any XFoil keystroke inputs that you'd like. By default, you start off within the OPER
        menu. All of the inputs indicated in the constructor have been set already, but you can override them here (for
        this run only) if you want.

        Returns: A dictionary containing all converged solutions obtained with your inputs.

        """
        # Set up a temporary directory
        with tempfile.TemporaryDirectory() as directory:
            directory = Path(directory)

            ### Alternatively, work in another directory:
            if self.working_directory is not None:
                directory = Path(self.working_directory)  # For debugging

            # Designate an intermediate file for file I/O
            output_filename = "output.txt"

            # Handle the airfoil file
            airfoil_file = "airfoil.dat"
            self.airfoil.write_dat(directory / airfoil_file)

            # Handle the keystroke file
            keystrokes = self._default_keystrokes(
                airfoil_filename=airfoil_file,
                output_filename=output_filename
            )
            keystrokes += [run_command]
            keystrokes += [
                "pacc",  # End polar accumulation
                "",
                "quit"
            ]

            # Remove an old output file, if one exists:
            try:
                os.remove(directory / output_filename)
            except FileNotFoundError:
                pass

            ### Execute
            try:
                # command = f'{self.xfoil_command} {airfoil_file}' # Old syntax; try this if calls are not working
                proc = subprocess.Popen(
                    self.xfoil_command,
                    cwd=directory,
                    stdin=subprocess.PIPE,
                    stdout=None if self.verbose else subprocess.DEVNULL,
                    stderr=None if self.verbose else subprocess.DEVNULL,
                    text=True,
                    # shell=True,
                    # timeout=self.timeout,
                    # check=True
                )
                outs, errs = proc.communicate(
                    input="\n".join(keystrokes),
                    timeout=self.timeout
                )
                return_code = proc.poll()

            except subprocess.TimeoutExpired:
                proc.kill()
                outs, errs = proc.communicate()

                warnings.warn(
                    "XFoil run timed out!\n"
                    "If this was not expected, try increasing the `timeout` parameter\n"
                    "when you create this AeroSandbox XFoil instance.",
                    stacklevel=2
                )
            except subprocess.CalledProcessError as e:
                if e.returncode == 11:
                    raise self.XFoilError(
                        "XFoil segmentation-faulted. This is likely because your input airfoil has too many points.\n"
                        "Try repaneling your airfoil with `Airfoil.repanel()` before passing it into XFoil.\n"
                        "For further debugging, turn on the `verbose` flag when creating this AeroSandbox XFoil instance.")
                elif e.returncode == 8 or e.returncode == 136:
                    raise self.XFoilError(
                        "XFoil returned a floating point exception. This is probably because you are trying to start\n"
                        "your analysis at an operating point where the viscous boundary layer can't be initialized based\n"
                        "on the computed inviscid flow. (You're probably hitting a Goldstein singularity.) Try starting\n"
                        "your XFoil run at a less-aggressive (alpha closer to 0, higher Re) operating point.")
                elif e.returncode == 1:
                    raise self.XFoilError(
                        f"Command '{self.xfoil_command}' returned non-zero exit status 1.\n"
                        f"This is likely because AeroSandbox does not see XFoil on PATH with the given command.\n"
                        f"Check the logs (`asb.XFoil(..., verbose=True)`) to verify that this is the case, and if so,\n"
                        f"provide the correct path to the XFoil executable in the asb.XFoil constructor via `xfoil_command=`."
                    )
                else:
                    raise e

            ### Parse the polar
            try:
                with open(directory / output_filename) as f:
                    lines = f.readlines()
            except FileNotFoundError:
                raise self.XFoilError(
                    "It appears XFoil didn't produce an output file, probably because it crashed.\n"
                    "To troubleshoot, try some combination of the following:\n"
                    "\t - In the XFoil constructor, verify that either XFoil is on PATH or that the `xfoil_command` parameter is set.\n"
                    "\t - In the XFoil constructor, run with `verbose=True`.\n"
                    "\t - In the XFoil constructor, set the `working_directory` parameter to a known folder to see the XFoil input and output files.\n"
                    "\t - In the XFoil constructor, set the `timeout` parameter to a large number to see if XFoil is just taking a long time to run.\n"
                    "\t - On Windows, use `XFoil.open_interactive()` to run XFoil interactively in a new window.\n"
                    "\t - Try allowing XFoil to repanel the airfoil by setting `xfoil_repanel=True` in the XFoil constructor.\n"
                )

            try:
                separator_line = None
                for i, line in enumerate(lines):
                    # The first line with at least 30 "-" in it is the separator line.
                    if line.count("-") >= 30:
                        separator_line = i
                        break

                if separator_line is None:
                    raise IndexError

                title_line = lines[i - 1]
                columns = title_line.split()

                data_lines = lines[i + 1:]

            except IndexError:
                raise self.XFoilError(
                    "XFoil output file is malformed; it doesn't have the expected number of lines.\n"
                    "For debugging, the raw output file from XFoil is printed below:\n"
                    + "\n".join(lines)
                    + "\nTitle line: " + title_line
                    + "\nColumns: " + str(columns)
                )

            def str_to_float(s: str) -> float:
                try:
                    return float(s)
                except ValueError:
                    return np.nan

            output = {
                column: []
                for column in [
                    "alpha",
                    "CL",
                    "CD",
                    "CDp",
                    "CM",
                    "Cpmin",
                    "Xcpmin",
                    "Chinge",
                    "Top_Xtr",
                    "Bot_Xtr",
                ]
            }

            for pointno, line in enumerate(data_lines):
                float_pattern = r"-?\d+\.\d+"
                entries = re.findall(float_pattern, line)
                data = [str_to_float(entry) for entry in entries]

                if len(data) == 10 and len(columns) == 8:
                    # This is a monkey-patch for a bug in XFoil v6.99, which causes polar output files to be malformed
                    # when including both Cpmin ("cinc") and hinge moment ("hinc") in the same run.
                    columns = [
                        "alpha",
                        "CL",
                        "CD",
                        "CDp",
                        "CM",
                        "Cpmin",
                        "Xcpmin",
                        "Chinge",
                        "Top_Xtr",
                        "Bot_Xtr",
                    ]

                if not len(data) == len(columns):
                    raise self.XFoilError(
                        "XFoil output file is malformed; the header and data have different numbers of columns.\n"
                        "In previous testing, this occurs due to a bug in XFoil itself, with certain input combos.\n"
                        "For debugging, the raw output file from XFoil is printed below:\n"
                        + "\n".join(lines)
                        + "\nTitle line: " + title_line
                        + f"\nIdentified {len(data)} data columns and {len(columns)} header columns."
                        + "\nColumns: " + str(columns)
                        + "\nData: " + str(data)
                    )

                for i in range(len(columns)):
                    output[columns[i]].append(data[i])

            output = {
                k: np.array(v, dtype=float)
                for k, v in output.items()
            }

            # Read the BL data
            if read_bl_data_from is not None:
                import pandas as pd
                bl_datas: List[pd.DataFrame] = []

                if read_bl_data_from == "alpha":
                    alpha_to_dump_mapping = {
                        float(dump_filename.stem.split("_")[-1]): dump_filename
                        for dump_filename in directory.glob("dump_a_*.txt")
                    }

                    for alpha in output["alpha"]:
                        dump_filename = alpha_to_dump_mapping[
                            min(alpha_to_dump_mapping.keys(), key=lambda x: abs(x - alpha))
                        ]

                        bl_datas.append(
                            pd.read_csv(
                                dump_filename,
                                sep=r"\s+",
                                names=["s", "x", "y", "ue/vinf", "dstar", "theta", "cf", "H"],
                                skiprows=1,
                            )
                        )

                elif read_bl_data_from == "cl":
                    cl_to_dump_mapping = {
                        float(dump_filename.stem.split("_")[-1]): dump_filename
                        for dump_filename in directory.glob("dump_cl_*.txt")
                    }

                    for cl in output["CL"]:
                        dump_filename = cl_to_dump_mapping[
                            min(cl_to_dump_mapping.keys(), key=lambda x: abs(x - cl))
                        ]

                        bl_datas.append(
                            pd.read_csv(
                                dump_filename,
                                sep=r"\s+",
                                names=["s", "x", "y", "ue/vinf", "dstar", "theta", "cf", "H"],
                                skiprows=1,
                            )
                        )

                else:
                    raise ValueError("The `read_bl_data_from` parameter must be 'alpha', 'cl', or None.")

                # Augment the output data for each BL
                for bl_data in bl_datas:
                    # Get Cp via Karman-Tsien compressibility correction, same as XFoil
                    Cp_0 = (1 - bl_data["ue/vinf"] ** 2)
                    bl_data["Cp"] = (Cp_0 /
                                     (
                                             np.sqrt(1 - self.mach ** 2)
                                             + (
                                                     (self.mach ** 2)
                                                     / (1 + np.sqrt(1 - self.mach ** 2))
                                                     * (Cp_0 / 2)
                                             )

                                     )
                                     )

                    # Get Re_theta
                    bl_data["Re_theta"] = np.abs(bl_data["ue/vinf"]) * bl_data["theta"] * self.Re

                output["bl_data"] = np.fromiter(bl_datas, dtype="O")

            return output

    def open_interactive(self) -> None:
        """
        Opens a new terminal window and runs XFoil interactively. This is useful for detailed analysis or debugging.

        Returns: None
        """
        with tempfile.TemporaryDirectory() as directory:
            directory = Path(directory)

            ### Alternatively, work in another directory:
            if self.working_directory is not None:
                directory = Path(self.working_directory)  # For debugging

            ### Handle the airplane file
            airfoil_file = "airfoil.dat"
            self.airfoil.write_dat(directory / airfoil_file)

            ### Open up AVL
            import sys, os
            if sys.platform == "win32":
                # Run XFoil
                print("Running XFoil interactively in a new window, quit it to continue...")

                command = f'cmd /k "{self.xfoil_command} {airfoil_file}"'

                process = subprocess.Popen(
                    command,
                    cwd=directory,
                    creationflags=subprocess.CREATE_NEW_CONSOLE
                )
                process.wait()

            else:
                raise NotImplementedError(
                    "Ability to auto-launch interactive XFoil sessions isn't yet implemented for non-Windows OSes."
                )

    def alpha(self,
              alpha: Union[float, np.ndarray],
              start_at: Union[float, None] = 0,
              ) -> Dict[str, np.ndarray]:
        """
        Execute XFoil at a given angle of attack, or at a sequence of angles of attack.

        Args:

            alpha: The angle of attack [degrees]. Can be either a float or an iterable of floats, such as an array.

            start_at: Chooses whether to split a large sweep into two runs that diverge away from some central value,
            to improve convergence. As an example, if you wanted to sweep from alpha=-20 to alpha=20, you might want
            to instead do two sweeps and stitch them together: 0 to 20, and 0 to -20. `start_at` can be either:

                * None, in which case the alpha inputs are run as a single sequence in the order given.

                * A float that corresponds to an angle of attack (in degrees), in which case the alpha inputs are
                split into two sequences that diverge from the `start_at` value. Successful runs are then sorted by
                `alpha` before returning.

        Returns: A dictionary with the XFoil results. Dictionary values are arrays; they may not be the same shape as
        your input array if some points did not converge.

        """
        alphas = np.reshape(np.array(alpha), -1)
        alphas = np.sort(alphas)

        commands = []

        def schedule_run(alpha: float):

            commands.append(f"a {alpha}")

            if self.hinge_point_x is not None:
                commands.append("fmom")

            if self.include_bl_data:
                commands.extend([
                    f"dump dump_a_{alpha:.8f}.txt",
                    # "vplo",
                    # "cd", # Dissipation coefficient
                    # f"dump cdis_a_{alpha:.8f}.txt",
                    # f"n", # Amplification ratio
                    # f"dump n_a_{alpha:.8f}.txt",
                    # "",
                ])

        if (
                len(alphas) > 1 and
                (start_at is not None) and
                (np.min(alphas) < start_at < np.max(alphas))
        ):
            alphas_upper = alphas[alphas > start_at]
            alphas_lower = alphas[alpha <= start_at][::-1]

            for a in alphas_upper:
                schedule_run(a)

            commands.append("init")

            for a in alphas_lower:
                schedule_run(a)
        else:
            for a in alphas:
                schedule_run(a)

        output = self._run_xfoil(
            "\n".join(commands),
            read_bl_data_from="alpha" if self.include_bl_data else None
        )

        sort_order = np.argsort(output['alpha'])
        output = {
            k: v[sort_order]
            for k, v in output.items()
        }
        return output

    def cl(self,
           cl: Union[float, np.ndarray],
           start_at: Union[float, None] = 0,
           ) -> Dict[str, np.ndarray]:
        """
        Execute XFoil at a given lift coefficient, or at a sequence of lift coefficients.

        Args:
            cl: The lift coefficient [-]. Can be either a float or an iterable of floats, such as an array.

            start_at: Chooses whether to split a large sweep into two runs that diverge away from some central value,
            to improve convergence. As an example, if you wanted to sweep from cl=-1.5 to cl=1.5, you might want to
            instead do two sweeps and stitch them together: 0 to 1.5, and 0 to -1.5. `start_at` can be either:

                * None, in which case the cl inputs are run as a single sequence in the order given.

                * A float that corresponds to an lift coefficient, in which case the cl inputs are
                split into two sequences that diverge from the `start_at` value. Successful runs are then sorted by
                `alpha` before returning.


        Returns: A dictionary with the XFoil results. Dictionary values are arrays; they may not be the same shape as
        your input array if some points did not converge.

        """
        cls = np.reshape(np.array(cl), -1)
        cls = np.sort(cls)

        commands = []

        def schedule_run(cl: float):

            commands.append(f"cl {cl}")

            if self.hinge_point_x is not None:
                commands.append("fmom")

            if self.include_bl_data:
                commands.extend([
                    f"dump dump_cl_{cl:.8f}.txt",
                    # "vplo",
                    # "cd", # Dissipation coefficient
                    # f"dump cdis_cl_{cl:.8f}.txt",
                    # f"n", # Amplification ratio
                    # f"dump n_cl_{cl:.8f}.txt",
                    # "",
                ])

        if (
                len(cls) > 1 and
                (start_at is not None) and
                (np.min(cls) < start_at < np.max(cls))
        ):
            cls_upper = cls[cls > start_at]
            cls_lower = cls[cls <= start_at][::-1]

            for c in cls_upper:
                schedule_run(c)

            commands.append("init")

            for c in cls_lower:
                schedule_run(c)
        else:
            for c in cls:
                schedule_run(c)

        output = self._run_xfoil(
            "\n".join(commands),
            read_bl_data_from="cl" if self.include_bl_data else None
        )

        sort_order = np.argsort(output['alpha'])
        output = {
            k: v[sort_order]
            for k, v in output.items()
        }
        return output


if __name__ == '__main__':
    af = Airfoil("naca2412").repanel(n_points_per_side=100)

    xf = XFoil(
        airfoil=af,
        Re=1e6,
        hinge_point_x=0.75,
        include_bl_data=True,
        working_directory=str(Path.home() / "Downloads" / "test"),
    )

    result_at_single_alpha = xf.alpha(5)

    result_at_several_CLs = xf.cl([-0.1, 0.5, 0.7, 0.8, 0.9])

    result_at_multiple_alphas = xf.alpha([3, 5, 60])
    # Note: if a result does not converge (such as the 60 degree case here), it will not be included in the results.



"""
A set of tools used for making prettier Matplotlib plots.
"""

### General imports
import seaborn as sns

### Local imports
from .plots import *
from .formatting import *
from .colors import *
from .annotation import *
from .threedim import *
from .quickplot import *

sns.set_theme(
    palette=palettes["categorical"],
)

mpl.rcParams["figure.dpi"] = 200
mpl.rcParams["axes.formatter.useoffset"] = False
mpl.rcParams["contour.negative_linestyle"] = 'solid'



import aerosandbox.numpy as np
from scipy import interpolate, integrate
import matplotlib.pyplot as plt
import aerosandbox.tools.pretty_plots as p
import sympy as s


def func_num(x, softness=1):
    return np.softmax(
        0,
        x,
        softness=softness
    )


def func_sym(x, softness=1):
    return s.log(
        1 + s.exp(
            x / softness
        )
    ) * softness
    # return s.sqrt(x ** 2 + softness ** 2)



softness = s.Rational(1, 100)

x = np.concatenate([
    np.sinspace(0, 200 * softness, 1000)[::-1][:-1] * -1,
    np.sinspace(0, 200 * softness, 1000),
]).astype(float)

# Discrete chirp function
f = func_num(x, softness=float(softness))
#
# f_interp = interpolate.InterpolatedUnivariateSpline(x, f, k=3)
# exact = integrate.quad(
#     lambda x: f_interp.derivative(2)(x) ** 2,
#     x[0], x[-1],
#     epsrel=1e-6
# )[0]

x_sym = s.symbols("x")
f_sym = func_sym(x_sym, softness=softness)
exact = s.integrate(
    f_sym.diff(x_sym, 2) ** 2,
    (x_sym, -s.oo, s.oo)
)

print(f"Exact: {exact}")

# Estimate discrete
slopes = np.diff(f) / np.diff(x)

subgradients = np.diff(slopes)

discrete = np.sum(subgradients ** 2)

print(f"Discrete: {discrete}")

print(f"Ratio: {discrete / exact} (equiv: 1 / {exact / discrete})")



import aerosandbox as asb
import aerosandbox.numpy as np
from typing import Callable, Union, Dict


class TubeSparBendingStructure(asb.ImplicitAnalysis):

    @asb.ImplicitAnalysis.initialize
    def __init__(self,
                 length: float,
                 diameter_function: Union[float, Callable[[np.ndarray], np.ndarray]] = None,
                 wall_thickness_function: Union[float, Callable[[np.ndarray], np.ndarray]] = None,
                 bending_point_forces: Dict[float, float] = None,
                 bending_distributed_force_function: Union[float, Callable[[np.ndarray], np.ndarray]] = 0.,
                 points_per_point_load: int = 20,
                 elastic_modulus_function: Union[float, Callable[[np.ndarray], np.ndarray]] = 175e9,  # Pa
                 EI_guess: float = None,
                 assume_thin_tube=True,
                 ):
        """
        A structural spar model that simulates bending of a cantilever tube spar based on beam theory (static,
        linear elasticity). This tube spar is assumed to have uniform wall thickness in the azimuthal direction,
        but not necessarily along its length. The diameter of the tube spar and elastic modulus may vary along its
        length.

        Governing equation is Euler-Bernoulli beam theory:

        (E * I * u(y)'')'' = q(y)

        where:
            * y is the distance along the spar, with a cantilever support at y=0 and a free tip at y=length.
            * E is the elastic modulus
            * I is the bending moment of inertia
            * u(y) is the local displacement at y.
            * q(y) is the force-per-unit-length at y. (In other words, a dirac delta is a point load.)
            * ()' is a derivative w.r.t. y.

        Any applicable constraints relating to stress, buckling, ovalization, gauge limits, displacement, etc. should
        be applied after initialization of this class.

        Example:

            >>> opti = asb.Opti()
            >>>
            >>> span = 34
            >>> half_span = span / 2
            >>> lift = 200 * 9.81
            >>>
            >>> beam = TubeSparBendingStructure(
            >>>     opti=opti,
            >>>     length=half_span,
            >>>     diameter_function=0.12,
            >>>     points_per_point_load=100,
            >>>     bending_distributed_force_function=lambda y: (lift / span) * (
            >>>             4 / np.pi * (1 - (y / half_span) ** 2) ** 0.5
            >>>     ),  # Elliptical
            >>>     # bending_distributed_force_function=lambda y: lift / span * np.ones_like(y) # Uniform
            >>> )
            >>> opti.subject_to([
            >>>     beam.stress_axial <= 500e6,  # Stress constraint
            >>>     beam.u[-1] <= 3,  # Tip displacement constraint
            >>>     beam.wall_thickness > 1e-3  # Gauge constraint
            >>> ])
            >>> mass = beam.volume() * 1600  # Density of carbon fiber [kg/m^3]
            >>>
            >>> opti.minimize(mass / 100)
            >>> sol = opti.solve()
            >>>
            >>> beam = sol(beam)
            >>>
            >>> print(f"{sol(mass)} kg")
            >>>
            >>> beam.draw()

        Args:

            length: Length of the spar [m]. Spar is assumed to go from y=0 (cantilever support) to y=length (free tip).

            diameter_function: The diameter of the tube as a function of the distance along the spar y. Refers to the
                nominal diameter (e.g., the arithmetic mean of the inner diameter and outer diameter of the tube; the
                "centerline" diameter). In terms of data types, this can be one of:

                * None, in which case it's interpreted as a design variable to optimize over. Assumes that the value
                can freely vary along the length of the spar.

                * a scalar optimization variable (see asb.ImplicitAnalysis documentation to see how to link an Opti
                instance to this analysis), in which case it's interpreted as a design variable to optimize over
                that's uniform along the length of the spar.

                * a float, in which case it's interpreted as a uniform value along the spar

                * a function (or other callable) in the form f(y), where y is the coordinate along the length of the
                spar. This function should be vectorized (e.g., a vector input of y values produces a vector output).

            wall_thickness_function: The wall thickness of the tube as a function of the distance along the spar y. In
                terms of data types, this can be one of:

                * None, in which case it's interpreted as a design variable to optimize over. Assumes that the value
                can freely vary along the length of the spar.

                * a scalar optimization variable (see asb.ImplicitAnalysis documentation to see how to link an Opti
                instance to this analysis), in which case it's interpreted as a design variable to optimize over
                that's uniform along the length of the spar.

                * a float, in which case it's interpreted as a uniform value along the spar

                * a function (or other callable) in the form f(y), where y is the coordinate along the length of the
                spar. This function should be vectorized (e.g., a vector input of y values produces a vector output).

            bending_point_forces: Not yet implemented; will allow for inclusion of point loads in the future.

            bending_distributed_force_function: The (distributed) load per unit span applied to the spar,
                as a function of the distance along the spar y. Should be in units of force per unit length. In terms of
                data types, this can be one of:

                * None, in which case it's interpreted as a design variable to optimize over. Assumes that the value
                can freely vary along the length of the spar.

                * a scalar optimization variable (see asb.ImplicitAnalysis documentation to see how to link an Opti
                instance to this analysis), in which case it's interpreted as a design variable to optimize over
                that's uniform along the length of the spar.

                * a float, in which case it's interpreted as a uniform value along the spar

                * a function (or other callable) in the form f(y), where y is the coordinate along the length of the
                spar. This function should be vectorized (e.g., a vector input of y values produces a vector output).

            points_per_point_load: Controls the discretization resolution of the beam. [int] When point load support
                is added, this will be the number of nodes between each individual point load.

            elastic_modulus_function: The elastic modulus [Pa] of the spar as a function of the distance along the
                spar y. In terms of data types, can be one of:

                * None, in which case it's interpreted as a design variable to optimize over. Assumes that the value
                can freely vary along the length of the spar.

                * a scalar optimization variable (see asb.ImplicitAnalysis documentation to see how to link an Opti
                instance to this analysis), in which case it's interpreted as a design variable to optimize over
                that's uniform along the length of the spar.

                * a float, in which case it's interpreted as a uniform value along the spar

                * a function (or other callable) in the form f(y), where y is the coordinate along the length of the
                spar. This function should be vectorized (e.g., a vector input of y values produces a vector output).

            EI_guess: Provides an initial guess for the bending stiffness EI, which is used in problems where spar
                diameter and thickness is not known at the outset. If not provided, a heuristic will be used to calculate this.

            assume_thin_tube: Makes assumptions that are applicable in the limit of a thin-walled (wall_thickness <<
                diameter) tube. This greatly increases numerical stability.

                Relative error of this assumption in the thin-walled limit is:

                    (wall_thickness / diameter) ^ 2

                So, for t/d = 0.1, the relative error is roughly 1%.

        """
        ### Parse the inputs
        self.length = length
        self.diameter_function = diameter_function
        self.wall_thickness_function = wall_thickness_function

        if bending_point_forces is not None:
            self.bending_point_forces = bending_point_forces
            raise NotImplementedError
        else:
            self.bending_point_forces = dict()

        self.bending_distributed_force_function = bending_distributed_force_function
        self.points_per_point_load = points_per_point_load
        self.elastic_modulus_function = elastic_modulus_function

        if EI_guess is None:
            try:
                diameter_guess = float(diameter_function)
            except (TypeError, RuntimeError):
                diameter_guess = 1

            try:
                wall_thickness_guess = float(wall_thickness_function)
            except (TypeError, RuntimeError):
                wall_thickness_guess = 0.01

            try:
                E_guess = float(elastic_modulus_function)
            except (TypeError, RuntimeError):
                E_guess = 175e9

            if assume_thin_tube:
                I_guess = np.pi / 8 * diameter_guess ** 3 * wall_thickness_guess
            else:
                I_guess = np.pi / 64 * (
                        (diameter_guess + wall_thickness_guess) ** 4 -
                        (diameter_guess - wall_thickness_guess) ** 4
                )
            EI_guess = E_guess * I_guess

            # EI_guess *= 1e0  # A very high EI guess is numerically stabilizing

        self.EI_guess = EI_guess

        self.assume_thin_tube = assume_thin_tube

        ### Discretize
        y = np.linspace(
            0,
            length,
            points_per_point_load
        )

        N = np.length(y)
        dy = np.diff(y)

        ### Evaluate the beam properties
        if isinstance(diameter_function, Callable):
            diameter = diameter_function(y)
        elif diameter_function is None:
            diameter = self.opti.variable(init_guess=1, n_vars=N, lower_bound=0.)
        else:
            diameter = diameter_function * np.ones_like(y)

        if isinstance(wall_thickness_function, Callable):
            wall_thickness = wall_thickness_function(y)
        elif wall_thickness_function is None:
            wall_thickness = self.opti.variable(init_guess=1e-2, n_vars=N, lower_bound=0, upper_bound=diameter)
        else:
            wall_thickness = wall_thickness_function * np.ones_like(y)

        if isinstance(bending_distributed_force_function, Callable):
            distributed_force = bending_distributed_force_function(y)
        else:
            distributed_force = bending_distributed_force_function * np.ones_like(y)

        if isinstance(elastic_modulus_function, Callable):
            elastic_modulus = elastic_modulus_function(y)
        else:
            elastic_modulus = elastic_modulus_function * np.ones_like(y)

        ### Evaluate the beam properties
        if assume_thin_tube:
            I = np.pi / 8 * diameter ** 3 * wall_thickness
        else:
            I = np.pi / 64 * (
                    (diameter + wall_thickness) ** 4 -
                    (diameter - wall_thickness) ** 4
            )
        EI = elastic_modulus * I

        ### Compute the initial guess
        u = self.opti.variable(
            init_guess=np.zeros_like(y),
            scale=np.sum(np.trapz(distributed_force) * dy) * length ** 4 / EI_guess
        )
        du = self.opti.derivative_of(
            u, with_respect_to=y,
            derivative_init_guess=np.zeros_like(y),
            derivative_scale=np.sum(np.trapz(distributed_force) * dy) * length ** 3 / EI_guess
        )
        ddu = self.opti.derivative_of(
            du, with_respect_to=y,
            derivative_init_guess=np.zeros_like(y),
            derivative_scale=np.sum(np.trapz(distributed_force) * dy) * length ** 2 / EI_guess
        )
        dEIddu = self.opti.derivative_of(
            EI * ddu, with_respect_to=y,
            derivative_init_guess=np.zeros_like(y),
            derivative_scale=np.sum(np.trapz(distributed_force) * dy) * length
        )
        self.opti.constrain_derivative(
            variable=dEIddu, with_respect_to=y,
            derivative=distributed_force
        )

        self.opti.subject_to([
            u[0] == 0,
            du[0] == 0,
            ddu[-1] == 0,
            dEIddu[-1] == 0
        ])

        bending_moment = -EI * ddu
        shear_force = -dEIddu

        stress_axial = elastic_modulus * ddu * (diameter + wall_thickness) / 2

        self.y = y
        self.diameter = diameter
        self.wall_thickness = wall_thickness
        self.distributed_force = distributed_force
        self.elastic_modulus = elastic_modulus
        self.I = I
        self.u = u
        self.du = du
        self.ddu = ddu
        self.dEIddu = dEIddu
        self.bending_moment = bending_moment
        self.shear_force = shear_force
        self.stress_axial = stress_axial

    def volume(self):
        if self.assume_thin_tube:
            return np.sum(
                np.pi * np.trapz(
                    self.diameter * self.wall_thickness
                ) * np.diff(self.y)
            )
        else:
            return np.sum(
                np.pi / 4 * np.trapz(
                    (self.diameter + self.wall_thickness) ** 2 -
                    (self.diameter - self.wall_thickness) ** 2
                ) * np.diff(self.y)
            )

    def total_force(self):
        if len(self.bending_point_forces) != 0:
            raise NotImplementedError

        return np.sum(
            np.trapz(
                self.distributed_force
            ) * np.diff(self.y)
        )

    def draw(self, show=True):
        import matplotlib.pyplot as plt
        import aerosandbox.tools.pretty_plots as p

        plot_quantities = {
            "Displacement [m]"              : self.u,
            # "Local Slope [deg]": np.arctan2d(self.du, 1),
            "Local Load [N/m]"              : self.distributed_force,
            "Axial Stress [MPa]"            : self.stress_axial / 1e6,
            "Bending $EI$ [N $\cdot$ m$^2$]": self.elastic_modulus * self.I,
            "Tube Diameter [m]"             : self.diameter,
            "Wall Thickness [m]"            : self.wall_thickness,
        }

        fig, ax = plt.subplots(2, 3, figsize=(8, 6), sharex='all')

        for i, (k, v) in enumerate(plot_quantities.items()):
            plt.sca(ax.flatten()[i])
            plt.plot(
                self.y,
                v,
                # ".-"
            )
            plt.ylabel(k)
            plt.xlim(
                np.min(self.y),
                np.max(self.y),
            )

        for a in ax[-1, :]:
            a.set_xlabel(r"$y$ [m]")

        if show:
            p.show_plot("Tube Spar Bending Structure")


if __name__ == '__main__':
    import aerosandbox.tools.units as u

    opti = asb.Opti()

    span = 112 * u.foot
    lift = 229 * u.lbm * 9.81
    half_span = span / 2

    beam = TubeSparBendingStructure(
        opti=opti,
        length=half_span,
        diameter_function=3.5 * u.inch,  # lambda y: (3.5 * u.inch) - (3.5 - 1.25) * u.inch * (y / half_span),
        points_per_point_load=100,
        bending_distributed_force_function=lambda y: (lift / span) * (
                4 / np.pi * (1 - (y / half_span) ** 2) ** 0.5
        ),  # Elliptical
        # bending_distributed_force_function=lambda y: lift / span * np.ones_like(y) # Uniform,
        elastic_modulus_function=228e9,
    )
    opti.subject_to([
        beam.stress_axial <= 500e6,  # Stress constraint
        beam.u[-1] <= 2,  # Tip displacement constraint
        beam.wall_thickness > 0.1e-3  # Gauge constraint
    ])
    mass = beam.volume() * 1600  # Density of carbon fiber [kg/m^3]

    opti.minimize(mass / (lift / 9.81))
    sol = opti.solve()

    beam = sol(beam)

    print(f"{sol(mass)} kg per half-wing")

    beam.draw()

    computed_spar_mass = 2 * sol(mass)

    vehicle_mass = lift / 9.81
    ultimate_load_factor = 2

    cruz_estimated_spar_mass = (
            (span * 1.17e-1 + span ** 2 * 1.10e-2) *
            (1 + (ultimate_load_factor * vehicle_mass / 100 - 2) / 4)
    )



import aerosandbox as asb
import aerosandbox.numpy as np
import pytest
from typing import Type

def test_wing_aero_3D_matches_2D_in_high_AR_limit():
    airfoil = asb.Airfoil("naca2412")

    wing = asb.Wing(
        xsecs=[
            asb.WingXSec(
                xyz_le=np.array([0, 0, 0]),
                chord=1,
                airfoil=airfoil,
            ),
            asb.WingXSec(
                xyz_le=np.array([0, 0.5e12, 0]),
                chord=1,
                airfoil=airfoil,
            )
        ]
    )

    airplane = asb.Airplane(
        wings=[wing],
    )

    Q = 1000

    op_point = asb.OperatingPoint(
        velocity=(2 * Q / 1.225) ** 0.5,
        alpha=5.,
    )

    xyz_ref = np.array([-1, 0, 0])

    ### Do 2D predictions
    airfoil_only_aero = airfoil.get_aero_from_neuralfoil(
        alpha=op_point.alpha,
        Re=op_point.reynolds(reference_length=1),
        mach=op_point.mach(),
        model_size="large"
    )
    expected_aero = {
        "CL": airfoil_only_aero["CL"],
        "CD": airfoil_only_aero["CD"],
        "Cm": (
                airfoil_only_aero["CM"]
                - (0.25 - xyz_ref[0]) * airfoil_only_aero["CL"]
        ),
    }

    ### Do 3D predictions
    aero = asb.AeroBuildup(
        airplane=airplane,
        op_point=op_point,
        xyz_ref=xyz_ref,
        model_size="large"
    ).run()

    ### Compare
    print(expected_aero)
    print({k: v for k, v in aero.items() if k in expected_aero})

    rtol = 0.05

    assert aero["CL"] == pytest.approx(expected_aero["CL"], rel=rtol)

    # assert aero["CD"] == pytest.approx(expected_aero["CD"], rel=rtol)
    """
    Drag convergence in the high-AR limit is NOT checked, which may appear surprising.
    
    The classical "intro-level" equation for induced drag, C_D_i = C_L^2 / (pi e AR), implies that induced drag goes to
    zero as the aspect ratio goes to infinity. 
    
    However, more advanced text consistently espouse a more nuanced view, where the induced drag actually does *not* 
    go to zero in the high-AR limit. (It does decrease, but the limit behavior is not exactly zero.) This is because 
    the induced drag is a function of the Oswald's efficiency (span efficiency) factor, e, which is itself a function 
    of aspect ratio.
    
    I was initially surprised to learn this and somewhat skeptical, but this view actually appears supported by a 
    rather-large fraction of the world's most elite aerodynamicists. For example:
    
        * Obert, "Aerodynamic Design of Transport Aircraft", 2009
        * Kroo, "Aircraft Design: Synthesis and Analysis", 2001
        * Stinton, "The Design of the Aeroplane", 2001
        * Hoerner, "Fluid-Dynamic Drag", 1965
    
    The models for e (Oswald's efficiency factor) that are given by these aerodynamicists are (partially) a function 
    of aspect ratio, and based on their form, the denominator of the induced drag equation (pi e AR) does not go to 
    zero in the high-AR limit.
    
    For a review paper of this phenomenon, which is counterintuitive based on the classical "intro-level" equation,
    see:
    
        * "Estimating the Oswald Factor from Basic Aircraft Geometrical Parameters" by M. Nita, D. Scholz; Hamburg 
        Univ. of Applied Sciences, 2012. https://www.fzt.haw-hamburg.de/pers/Scholz/OPerA/OPerA_PUB_DLRK_12-09-10.pdf

    """

    assert aero["Cm"] == pytest.approx(expected_aero["Cm"], rel=rtol)


wing = asb.Wing(
    symmetric=True,
    xsecs=[
        asb.WingXSec(
            xyz_le=np.array([0., 0, 0]),
            chord=1.,
            airfoil=asb.Airfoil("naca2412"),
        ),
        asb.WingXSec(
            xyz_le=np.array([0., 5, 0]),
            chord=1.,
            airfoil=asb.Airfoil("naca2412"),
        ),
    ]
)

airplane = asb.Airplane(
    wings=[wing],
)


def test_simple_wing_stability_derivatives(
        AeroAnalysis: Type = asb.AeroBuildup,
):
    analysis = AeroAnalysis(
        airplane=airplane,
        op_point=asb.OperatingPoint(
            velocity=100.,
            alpha=0.,
            beta=5.,
        ),
        xyz_ref=np.array([
            0.,
            0.,
            0.
        ]),
    )

    try:
        aero = analysis.run_with_stability_derivatives()
    except AttributeError:
        aero = analysis.run()

    print(f"Aerodynamic coefficients with {AeroAnalysis.__name__}:")
    for key in ["CL", "CD", "CY", "Cl", "Cm", "Cn", "Cma"]:
        print(f"{key.rjust(10)}: {float(aero[key]):20.4f}")


if __name__ == '__main__':
    test_wing_aero_3D_matches_2D_in_high_AR_limit()
    test_simple_wing_stability_derivatives()



import sympy as s
from sympy import init_printing
init_printing()

# Gets the integral from x2 to x3, looking at the cubic spline interpolant from x1...x4

# Define the symbols
x1, x2, x3, x4 = s.symbols('x1 x2 x3 x4', real=True)
f1, f2, f3, f4 = s.symbols('f1 f2 f3 f4', real=True)

h = x3 - x2
hm = x2 - x1
hp = x4 - x3

q = s.symbols('q')  # Normalized space for a Bernstein basis.
# Mapping from x-space to q-space has x=x2 -> q=0, x=x3 -> q=1.

q2 = 0
q3 = 1
# q1 = q2 - hm / h
# q4 = q3 + hp / h
q1, q4 = s.symbols('q1 q4', real=True)

# Define the Bernstein basis polynomials
b1 = (1 - q) ** 3
b2 = 3 * q * (1 - q) ** 2
b3 = 3 * q ** 2 * (1 - q)
b4 = q ** 3

c1, c2, c3, c4 = s.symbols('c1 c2 c3 c4', real=True)

# Can solve for c2 and c3 exactly
c1 = f2
c4 = f3

f = c1 * b1 + c2 * b2 + c3 * b3 + c4 * b4

f1_cubic = f.subs(q, q1)#.simplify()
f4_cubic = f.subs(q, q4)#.simplify()

factors = [q1, q4]
# factors = [f1, f2, f3, f4]

# Solve for c2 and c3
sol = s.solve(
    [
        f1_cubic - f1,
        f4_cubic - f4,
    ],
    [
        c2,
        c3,
    ],
)
c2 = sol[c2].factor(factors).simplify()
c3 = sol[c3].factor(factors).simplify()



integral = (c1 + c2 + c3 + c4) / 4 # God I love Bernstein polynomials
# integral = s.simplify(integral)

integral = integral.factor(factors).simplify()

parsimony = len(str(integral))
print(s.pretty(integral, num_columns=100))
print(f"Parsimony: {parsimony}")


import numpy as _onp
from numpy import pi as _pi

_deg2rad = 180. / _pi
_rad2deg = _pi / 180.


def degrees(x):
    """Converts an input x from radians to degrees"""
    return x * _deg2rad


def radians(x):
    """Converts an input x from degrees to radians"""
    return x * _rad2deg


def sind(x):
    """Returns the sin of an angle x, given in degrees"""
    return _onp.sin(radians(x))


def cosd(x):
    """Returns the cos of an angle x, given in degrees"""
    return _onp.cos(radians(x))


def tand(x):
    """Returns the tangent of an angle x, given in degrees"""
    return _onp.tan(radians(x))


def arcsind(x):
    """Returns the arcsin of an x, in degrees"""
    return degrees(_onp.arcsin(x))


def arccosd(x):
    """Returns the arccos of an x, in degrees"""
    return degrees(_onp.arccos(x))


def arctan2d(y, x):
    """Returns the angle associated with arctan(y, x), in degrees"""
    return degrees(_onp.arctan2(y, x))



from pathlib import Path
import sys, os
from aerosandbox.tools.python.io import convert_ipynb_to_py


def run_python_file(path: Path) -> None:
    """
    Executes a Python file from a path.
    Args:
        path: File path

    Returns: None

    """
    sys.path.append(str(path.parent))
    __import__(path.with_suffix("").name)


def run_all_python_files(path: Path, recursive=True, verbose=True) -> None:
    """
    Executes all Python files and Jupyter Notebooks in a directory.
    Args:
        path: A Path-type object (Path from built-in pathlib) representing a filepath
        recursive: Executes recursively (e.g. searches all subfolders too)

    Returns: None

    """
    # Exclusions:
    if path == Path(os.path.abspath(__file__)):  # Don't run this file
        return
    if "ignore" in str(path).lower():  # Don't run any file or folder with the word "ignore" in the name.
        return

    if path.is_file():

        ### Run the file if it's a Python file
        if path.suffix == ".py":
            if verbose:
                print(f"##### Running file: {path}")
            run_python_file(path)

        ### Run the file if it's a Jupyter notebook
        if path.suffix == ".ipynb":
            if verbose:
                print(f"##### Converting file: {path}")
            notebook = path
            python_file = path.with_suffix(".py")

            convert_ipynb_to_py(notebook, python_file)

            run_all_python_files(python_file, recursive=False, verbose=verbose)

    elif path.is_dir() and recursive:
        if verbose:
            print(f"##### Opening directory: {path}")
        for subpath in path.iterdir():
            run_all_python_files(subpath, recursive=recursive, verbose=verbose)



import aerosandbox as asb
import aerosandbox.numpy as np
import pytest

def test_dyn_indexing():

    # Test indexing of a simple Dynamics object
    dyn = asb.DynamicsPointMass1DHorizontal(
        mass_props=asb.MassProperties(
            mass=1
        ),
        x_e=np.arange(10) ** 2,
        u_e=2 * np.arange(10),
    )

    d0 = dyn[0]
    d1 = dyn[1]
    dn1 = dyn[-1]
    dn2 = dyn[-2]
    dslice = dyn[2:5]

    assert d0.x_e == 0
    assert d0.u_e == 0
    assert d1.x_e == 1
    assert d1.u_e == 2
    assert dn1.x_e == 81
    assert dn1.u_e == 18
    assert dn2.x_e == 64
    assert dn2.u_e == 16
    assert all(dslice.x_e == [4, 9, 16])
    assert all(dslice.u_e == [4, 6, 8])

if __name__ == '__main__':
    test_dyn_indexing()


import aerosandbox.numpy as np
import warnings
from aerosandbox.library.aerodynamics.viscous import Cf_flat_plate

warnings.warn(
    "This file is deprecated and will be removed in the subsequent version of ASB.\n"
    "You can use `asb.Airfoil.get_aero_from_neuralfoil()` instead\n"
    "to get airfoil aerodynamics for any airfoil.",
    DeprecationWarning
)


def default_CL_function(alpha, Re, mach=0, deflection=0):
    """
    Lift coefficient.
    """
    Cl_inc = np.pi * np.sind(2 * alpha)
    beta = (1 - mach) ** 2

    Cl = Cl_inc * beta
    return Cl


def default_CD_function(alpha, Re, mach=0, deflection=0):
    """
    Drag coefficient.
    """
    Cf = Cf_flat_plate(Re_L=Re, method="hybrid-sharpe-convex")

    ### Form factor model from Raymer, "Aircraft Design". Section 12.5, Eq. 12.30
    t_over_c = 0.12
    FF = 1 + 2 * t_over_c * 100 * t_over_c ** 4

    Cd_inc = 2 * Cf * FF * (
            1 + (np.sind(alpha) * 180 / np.pi / 5) ** 2
    )
    beta = (1 - mach) ** 2

    Cd = Cd_inc * beta
    return Cd


def default_CM_function(alpha, Re, mach=0, deflection=0):
    """
    Pitching moment coefficient, as measured about quarter-chord.
    """
    return np.zeros_like(alpha)



import aerosandbox.numpy as np
from aerosandbox import ExplicitAnalysis, AeroSandboxObject
from aerosandbox.geometry import *
from aerosandbox.performance import OperatingPoint
from aerosandbox.aerodynamics.aero_3D.singularities.uniform_strength_horseshoe_singularities import \
    calculate_induced_velocity_horseshoe
from typing import Dict, Any, List, Callable, Optional, Union, Tuple
import copy
from functools import cached_property, lru_cache, partial
from collections import namedtuple
from dataclasses import dataclass
from abc import abstractmethod, ABC


### Define some helper functions that take a vector and make it a Nx1 or 1xN, respectively.
# Useful for broadcasting with matrices later.
def tall(array):
    return np.reshape(array, (-1, 1))


def wide(array):
    return np.reshape(array, (1, -1))


immutable_dataclass = partial(dataclass, frozen=True, repr=False)


class LinearPotentialFlow(ExplicitAnalysis):

    def __init__(self,
                 airplane: Airplane,
                 op_point: OperatingPoint,
                 xyz_ref: List[float] = None,
                 run_symmetric_if_possible: bool = False,
                 verbose: bool = False,
                 wing_model: Union[str, Dict[Wing, str]] = "vortex_lattice_all_horseshoe",
                 fuselage_model: Union[str, Dict[Fuselage, str]] = "none",
                 wing_options: Union[Dict[str, Any], Dict[Wing, Dict[str, Any]]] = None,
                 fuselage_options: Union[Dict[str, Any], Dict[Fuselage, Dict[str, Any]]] = None,
                 ):
        import warnings
        warnings.warn("LinearPotentialFlow is under active development and is not yet ready for use.", UserWarning)

        super().__init__()

        ##### Set defaults
        if xyz_ref is None:
            xyz_ref = airplane.xyz_ref
        if wing_options is None:
            wing_options = {}
        if fuselage_options is None:
            fuselage_options = {}

        ##### Initialize
        self.airplane = airplane
        self.op_point = op_point
        self.xyz_ref = xyz_ref
        self.verbose = verbose

        ##### Set up the modeling methods
        if isinstance(wing_model, str):
            wing_model = {wing: wing_model for wing in self.airplane.wings}
        if isinstance(fuselage_model, str):
            fuselage_model = {fuselage: fuselage_model for fuselage in self.airplane.fuselages}

        self.wing_model: Dict[Wing, str] = wing_model
        self.fuselage_model: Dict[Fuselage, str] = fuselage_model

        ##### Set up the modeling options
        ### Check the format of the wing options
        if not (
                all([isinstance(k, str) for k in wing_options.keys()]) or
                all([issubclass(k, Wing) for k in wing_options.keys()])
        ):
            raise ValueError("`wing_options` must be either:\n"
                             "    - A dictionary of the form `{str: value}`, which is applied to all Wings\n"
                             "    - A nested dictionary of the form `{Wing: {str: value}}`, which is applied to the corresponding Wings\n"
                             )
        elif all([isinstance(k, str) for k in wing_options.keys()]):
            wing_options = {wing: wing_options for wing in self.airplane.wings}

        ### Check the format of the fuselage options
        if not (
                all([isinstance(k, str) for k in fuselage_options.keys()]) or
                all([issubclass(k, Fuselage) for k in fuselage_options.keys()])
        ):
            raise ValueError("`fuselage_options` must be either:\n"
                             "    - A dictionary of the form `{str: value}`, which is applied to all Fuselages\n"
                             "    - A nested dictionary of the form `{Fuselage: {str: value}}`, which is applied to the corresponding Fuselages\n"
                             )
        elif all([isinstance(k, str) for k in fuselage_options.keys()]):
            fuselage_options = {fuselage: fuselage_options for fuselage in self.airplane.fuselages}

        ### Set user-specified values
        self.wing_options: Dict[Wing, Dict[str, Any]] = wing_options
        self.fuselage_options: Dict[Fuselage, Dict[str, Any]] = fuselage_options

        ### Set default values
        wing_model_default_options = {
            "none"                        : {},
            "vortex_lattice_all_horseshoe": {
                "spanwise_resolution"              : 10,
                "spanwise_spacing_function"        : np.cosspace,
                "chordwise_resolution"             : 10,
                "chordwise_spacing_function"       : np.cosspace,
                "vortex_core_radius"               : 1e-8,
                "align_trailing_vortices_with_wind": False,
            },
            "vortex_lattice_ring"         : {
                "spanwise_resolution"              : 10,
                "spanwise_spacing_function"        : np.cosspace,
                "chordwise_resolution"             : 10,
                "chordwise_spacing_function"       : np.cosspace,
                "vortex_core_radius"               : 1e-8,
                "align_trailing_vortices_with_wind": False,
            },
            "lifting_line"                : {
                "sectional_data_source": "neuralfoil",
            },
        }

        for wing in self.airplane.wings:
            if self.wing_model[wing] in wing_model_default_options.keys():
                self.wing_options[wing] = {
                    **wing_model_default_options[self.wing_model[wing]],
                    **self.wing_options[wing],
                }
            else:
                raise ValueError(f"Invalid wing model specified: \"{self.wing_model[wing]}\"\n"
                                 f"Must be one of: {list(wing_model_default_options.keys())}")

        fuselage_model_default_options = {
            "none"                  : {},
            "prescribed_source_line": {
                "lengthwise_resolution"      : 1,
                "lengthwise_spacing_function": np.cosspace,
            },
        }

        for fuselage in self.airplane.fuselages:
            if self.fuselage_model[fuselage] in fuselage_model_default_options.keys():
                self.fuselage_options[fuselage] = {
                    **fuselage_model_default_options[self.fuselage_model[fuselage]],
                    **self.fuselage_options[fuselage],
                }
            else:
                raise ValueError(f"Invalid fuselage model specified: \"{self.fuselage_model[fuselage]}\"\n"
                                 f"Must be one of: {list(fuselage_model_default_options.keys())}")

        ### Determine whether you should run the problem as symmetric
        self.run_symmetric = False
        if run_symmetric_if_possible:
            raise NotImplementedError("LinearPotentialFlow with symmetry detection not yet implemented!")
            # try:
            #     self.run_symmetric = (  # Satisfies assumptions
            #             self.op_point.beta == 0 and
            #             self.op_point.p == 0 and
            #             self.op_point.r == 0 and
            #             self.airplane.is_entirely_symmetric()
            #     )
            # except RuntimeError:  # Required because beta, p, r, etc. may be non-numeric (e.g. opti variables)
            #     pass

    def __repr__(self):
        return self.__class__.__name__ + "(\n" + "\n".join([
            f"\tairplane={self.airplane}",
            f"\top_point={self.op_point}",
            f"\txyz_ref={self.xyz_ref}",
        ]) + "\n)"

    @immutable_dataclass
    class Elements(ABC):
        parent_component: AeroSandboxObject
        start_index: int
        end_index: int

        def __repr__(self):
            return self.__class__.__name__ + "(\n" + "\n".join([
                f"\tparent_component={self.parent_component}",
                f"\tstart_index={self.start_index}",
                f"\tend_index={self.end_index}",
                f"\tlength={len(self)}",
            ]) + "\n)"

        @abstractmethod
        def __len__(self):
            pass

        # @abstractmethod
        # def get_induced_velocity_at_points(self,
        #                                    points: np.ndarray
        #                                    ) -> np.ndarray:
        #     pass

    @immutable_dataclass
    class PanelElements(Elements, ABC):
        front_left_vertices: np.ndarray  # Nx3 array of panel corners
        back_left_vertices: np.ndarray  # Nx3 array of panel corners
        back_right_vertices: np.ndarray  # Nx3 array of panel corners
        front_right_vertices: np.ndarray  # Nx3 array of panel corners

        def __len__(self):
            return np.length(self.front_left_vertices)

        @cached_property
        def crosses(self):
            diag1 = self.front_right_vertices - self.back_left_vertices
            diag2 = self.front_left_vertices - self.back_right_vertices
            return np.cross(diag1, diag2)

        @cached_property
        def cross_norms(self):
            return np.linalg.norm(self.crosses, axis=1)

        @cached_property
        def areas(self) -> np.ndarray:
            return self.cross_norms / 2

        @cached_property
        def normal_directions(self):
            return self.crosses / tall(self.cross_norms)

    @immutable_dataclass
    class WingHorseshoeVortexElements(PanelElements):
        trailing_vortex_direction: np.ndarray  # Nx3 array of trailing vortex directions
        vortex_core_radius: float  # [meters]

        @cached_property
        def left_vortex_vertices(self):
            return 0.75 * self.front_left_vertices + 0.25 * self.back_left_vertices

        @cached_property
        def right_vortex_vertices(self):
            return 0.75 * self.front_right_vertices + 0.25 * self.back_right_vertices

        @cached_property
        def vortex_centers(self):
            return (self.left_vortex_vertices + self.right_vortex_vertices) / 2

        @cached_property
        def vortex_bound_legs(self):
            return self.right_vortex_vertices - self.left_vortex_vertices

        @cached_property
        def collocation_points(self):
            return (
                    0.5 * (0.25 * self.front_left_vertices + 0.75 * self.back_left_vertices) +
                    0.5 * (0.25 * self.front_right_vertices + 0.75 * self.back_right_vertices)
            )

        def get_induced_velocity_at_points(self,
                                           points: np.ndarray,
                                           vortex_strengths: np.ndarray,
                                           sum_across_elements: bool = True
                                           ) -> Tuple[np.ndarray]:
            u_induced, v_induced, w_induced = calculate_induced_velocity_horseshoe(
                x_field=tall(points[:, 0]),
                y_field=tall(points[:, 1]),
                z_field=tall(points[:, 2]),
                x_left=wide(self.left_vortex_vertices[:, 0]),
                y_left=wide(self.left_vortex_vertices[:, 1]),
                z_left=wide(self.left_vortex_vertices[:, 2]),
                x_right=wide(self.right_vortex_vertices[:, 0]),
                y_right=wide(self.right_vortex_vertices[:, 1]),
                z_right=wide(self.right_vortex_vertices[:, 2]),
                trailing_vortex_direction=self.trailing_vortex_direction,
                gamma=wide(vortex_strengths),
                vortex_core_radius=self.vortex_core_radius
            )

            if sum_across_elements:
                u_induced = np.sum(u_induced, axis=1)
                v_induced = np.sum(v_induced, axis=1)
                w_induced = np.sum(w_induced, axis=1)

            return u_induced, v_induced, w_induced

    @immutable_dataclass
    class WingLiftingLineElements(WingHorseshoeVortexElements):
        CL0: np.ndarray  # length N array of zero-angle-of-attack lift coefficients
        CLa: np.ndarray  # length N array of lift slopes (1/rad)

        @property
        def collocation_points(self):
            raise NotImplementedError

    @immutable_dataclass
    class FuselagePrescribedSourceLineElements(Elements):
        front_vertices: np.ndarray
        back_vertices: np.ndarray
        strength: np.ndarray

        def __len__(self):
            return np.length(self.front_vertices)

    @cached_property
    def discretization(self):
        """

        Returns: A list of dictionaries, where each item in the list represents a single element.

            Each item in the list is a namedtuple (effectively, a dictionary), and one of the following types:

                    * `wing_vlm_element`
                    * `wing_lifting_line_element`
                    * `fuselage_prescribed_source_line`

        """
        ### Initialize
        discretization = []
        index = 0

        ### Wings
        for wing in self.airplane.wings:
            element_type: str = self.wing_model[wing]
            options = self.wing_options[wing]

            if element_type == "none":
                continue

            elif element_type == "vortex_lattice_all_horseshoe":
                if options["spanwise_resolution"] > 1:
                    subdivided_wing = wing.subdivide_sections(
                        ratio=options["spanwise_resolution"],
                        spacing_function=options["spanwise_spacing_function"],
                    )
                else:
                    subdivided_wing = wing

                points, faces = subdivided_wing.mesh_thin_surface(
                    method="quad",
                    chordwise_resolution=options["chordwise_resolution"],
                    chordwise_spacing_function=options["chordwise_spacing_function"],
                    add_camber=True
                )

                if options["align_trailing_vortices_with_wind"]:
                    raise NotImplementedError("align_trailing_vortices_with_wind not yet implemented!")
                else:
                    trailing_vortex_direction = np.array([1, 0, 0])

                discretization.append(
                    self.WingHorseshoeVortexElements(
                        parent_component=wing,
                        start_index=index,
                        end_index=(index := index + len(faces)),
                        front_left_vertices=points[faces[:, 0], :],
                        back_left_vertices=points[faces[:, 1], :],
                        back_right_vertices=points[faces[:, 2], :],
                        front_right_vertices=points[faces[:, 3], :],
                        trailing_vortex_direction=trailing_vortex_direction,
                        vortex_core_radius=options["vortex_core_radius"],
                    )
                )
            elif element_type == "vortex_lattice_ring":
                raise NotImplementedError("vortex_lattice_ring not yet implemented!")

            elif element_type == "lifting_line":
                raise NotImplementedError("lifting_line not yet implemented!")

            else:
                raise ValueError(f"Invalid wing model specified: \"{element_type}\"")

        ### Fuselages
        for fuselage in self.airplane.fuselages:
            element_type: str = self.fuselage_model[fuselage]
            options = self.fuselage_options[fuselage]

            if element_type == "none":
                continue

            elif element_type == "prescribed_source_line":
                raise NotImplementedError("prescribed_source_line not yet implemented!")

            else:
                raise ValueError(f"Invalid fuselage model specified: \"{element_type}\"")

        return discretization

    @cached_property
    def N_elements(self):
        return sum([len(element_collection) for element_collection in self.discretization])

    @cached_property
    def AIC(self):

        A = np.empty((self.N_elements, self.N_elements)) * np.nan

        for element_collection in self.discretization:
            if isinstance(element_collection, self.WingHorseshoeVortexElements):
                raise NotImplementedError("AIC not yet implemented for horseshoe vortices.")
            elif isinstance(element_collection, self.WingLiftingLineElements):
                raise NotImplementedError("AIC not yet implemented for lifting lines.")
            elif isinstance(element_collection, self.FuselagePrescribedSourceLineElements):
                raise NotImplementedError("AIC not yet implemented for fuselages.")
            else:
                raise ValueError(f"Invalid element type: {type(element_collection)}")

        return A

    def run(self) -> Dict[str, Any]:
        """
        Computes the aerodynamic forces.

        Returns a dictionary with keys:

            - 'F_g' : an [x, y, z] list of forces in geometry axes [N]
            - 'F_b' : an [x, y, z] list of forces in body axes [N]
            - 'F_w' : an [x, y, z] list of forces in wind axes [N]
            - 'M_g' : an [x, y, z] list of moments about geometry axes [Nm]
            - 'M_b' : an [x, y, z] list of moments about body axes [Nm]
            - 'M_w' : an [x, y, z] list of moments about wind axes [Nm]
            - 'L' : the lift force [N]. Definitionally, this is in wind axes.
            - 'Y' : the side force [N]. This is in wind axes.
            - 'D' : the drag force [N]. Definitionally, this is in wind axes.
            - 'l_b', the rolling moment, in body axes [Nm]. Positive is roll-right.
            - 'm_b', the pitching moment, in body axes [Nm]. Positive is pitch-up.
            - 'n_b', the yawing moment, in body axes [Nm]. Positive is nose-right.
            - 'CL', the lift coefficient [-]. Definitionally, this is in wind axes.
            - 'CY', the sideforce coefficient [-]. This is in wind axes.
            - 'CD', the drag coefficient [-]. Definitionally, this is in wind axes.
            - 'Cl', the rolling coefficient [-], in body axes
            - 'Cm', the pitching coefficient [-], in body axes
            - 'Cn', the yawing coefficient [-], in body axes

        Nondimensional values are nondimensionalized using reference values in the LinearPotentialFlow.airplane object.
        """

        raise NotImplementedError

    def get_induced_velocity_at_points(self,
                                       points: np.ndarray
                                       ) -> np.ndarray:
        raise NotImplementedError

    def get_velocity_at_points(self,
                               points: np.ndarray
                               ) -> np.ndarray:
        raise NotImplementedError

    def get_streamlines(self,
                        seed_points: np.ndarray = None,
                        n_steps: int = 300,
                        length: float = None,
                        ):
        raise NotImplementedError

    def draw(self,
             c: np.ndarray = None,
             cmap: str = None,
             colorbar_label: str = None,
             show: bool = True,
             show_kwargs: Dict = None,
             draw_streamlines=True,
             recalculate_streamlines=False,
             backend: str = "pyvista"
             ):
        raise NotImplementedError

    def draw_three_view(self):
        raise NotImplementedError


if __name__ == '__main__':
    ### Import Vanilla Airplane
    import aerosandbox as asb

    from pathlib import Path

    geometry_folder = Path(__file__).parent / "test_aero_3D" / "geometries"

    import sys

    sys.path.insert(0, str(geometry_folder))

    from vanilla import airplane as vanilla

    ### Do the AVL run
    lpf = LinearPotentialFlow(
        airplane=vanilla,
        op_point=asb.OperatingPoint(
            atmosphere=asb.Atmosphere(altitude=0),
            velocity=10,
            alpha=0,
            beta=0,
            p=0,
            q=0,
            r=0,
        ),
    )

    dis = lpf.discretization

    res = lpf.run()

    for k, v in res.items():
        print(f"{str(k).rjust(10)} : {v}")
